from unittest import mock
from unittest.mock import PropertyMock

import pytest

from solaar.ui import common


@pytest.mark.parametrize(
    "reason, expected_in_title, expected_in_text",
    [
        (
            common.ErrorReason.PERMISSIONS,
            "Permissions error",
            "not have permission to open",
        ),
        (common.ErrorReason.NO_DEVICE, "connect to device error", "error connecting"),
        (common.ErrorReason.UNPAIR, "Unpairing failed", "receiver returned an error"),
    ],
)
def test_create_error_text(reason, expected_in_title, expected_in_text):
    obj = mock.Mock()
    obj.name = PropertyMock(return_value="test")

    title, text = common._create_error_text(reason, obj)

    assert expected_in_title in title
    assert expected_in_text in text



from hid_parser.data import Button
from hid_parser.data import Consumer


def test_consumer():
    consumer = Consumer()

    assert consumer.PLAY_PAUSE == 0xCD


def test_button():
    button = Button()

    assert button.NO_BUTTON == 0x0
    assert button.BUTTON_1 == 0x1



## Copyright (C) 2024  Solaar Contributors https://pwr-solaar.github.io/Solaar/
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from dataclasses import dataclass
from functools import partial
from typing import Optional

import pytest

from logitech_receiver import common
from logitech_receiver import device
from logitech_receiver import hidpp20
from logitech_receiver.common import BatteryLevelApproximation
from logitech_receiver.common import BatteryStatus

from . import fake_hidpp


class LowLevelInterfaceFake:
    def __init__(self, responses=None):
        self.responses = responses

    def open_path(self, path) -> int:
        return fake_hidpp.open_path(path)

    def find_paired_node(self, receiver_path: str, index: int, timeout: int):
        return None

    def request(self, response, *args, **kwargs):
        func = partial(fake_hidpp.request, self.responses)
        return func(response, *args, **kwargs)

    def ping(self, response, *args, **kwargs):
        func = partial(fake_hidpp.ping, self.responses)
        return func(response, *args, **kwargs)

    def close(self, *args, **kwargs):
        pass


@dataclass
class DeviceInfoStub:
    path: str
    product_id: str
    vendor_id: int = 1133
    hidpp_short: bool = False
    hidpp_long: bool = True
    bus_id: int = 0x0003  # USB
    serial: str = "aa:aa:aa;aa"


di_bad_handle = DeviceInfoStub(None, product_id="CCCC")
di_error = DeviceInfoStub(11, product_id="CCCC")
di_CCCC = DeviceInfoStub("11", product_id="CCCC")
di_C318 = DeviceInfoStub("11", product_id="C318")
di_B530 = DeviceInfoStub("11", product_id="B350", bus_id=0x0005)
di_C068 = DeviceInfoStub("11", product_id="C06B")
di_C08A = DeviceInfoStub("11", product_id="C08A")
di_DDDD = DeviceInfoStub("11", product_id="DDDD")


@pytest.mark.parametrize(
    "device_info, responses, expected_success",
    [
        (di_bad_handle, fake_hidpp.r_empty, None),
        (di_error, fake_hidpp.r_empty, False),
        (di_CCCC, fake_hidpp.r_empty, True),
    ],
)
def test_create_device(device_info, responses, expected_success):
    low_level_mock = LowLevelInterfaceFake(responses)
    if expected_success is None:
        with pytest.raises(PermissionError):
            device.create_device(low_level_mock, device_info)
    elif not expected_success:
        with pytest.raises(Exception):  # noqa: B017
            device.create_device(low_level_mock, device_info)
    else:
        test_device = device.create_device(low_level_mock, device_info)
        assert bool(test_device) == expected_success


@pytest.mark.parametrize(
    "device_info, responses, expected_codename, expected_name, expected_kind",
    [(di_CCCC, fake_hidpp.r_empty, "?? (CCCC)", "Unknown device CCCC", "?")],
)
def test_device_name(device_info, responses, expected_codename, expected_name, expected_kind):
    low_level = LowLevelInterfaceFake(responses)

    test_device = device.create_device(low_level, device_info)

    assert test_device.codename == expected_codename
    assert test_device.name == expected_name
    assert test_device.kind == expected_kind


@pytest.mark.parametrize(
    "device_info, responses, handle, _name, _codename, number, protocol, registers",
    zip(
        [di_CCCC, di_C318, di_B530, di_C068, di_C08A, di_DDDD],
        [
            fake_hidpp.r_empty,
            fake_hidpp.r_keyboard_1,
            fake_hidpp.r_keyboard_2,
            fake_hidpp.r_mouse_1,
            fake_hidpp.r_mouse_2,
            fake_hidpp.r_mouse_3,
        ],
        [0x11, 0x11, 0x11, 0x11, 0x11, 0x11],
        [None, "Illuminated Keyboard", "Craft Advanced Keyboard", "G700 Gaming Mouse", "MX Vertical Wireless Mouse", None],
        [None, "Illuminated", "Craft", "G700", "MX Vertical", None],
        [0xFF, 0x0, 0xFF, 0x0, 0xFF, 0xFF],
        [1.0, 1.0, 4.5, 1.0, 4.5, 4.5],
        [[], [], [], (common.NamedInt(7, "battery status"), common.NamedInt(81, "three leds")), [], []],
    ),
)
def test_device_info(device_info, responses, handle, _name, _codename, number, protocol, registers):
    test_device = device.Device(LowLevelInterfaceFake(responses), None, None, None, handle=handle, device_info=device_info)

    assert test_device.handle == handle
    assert test_device._name == _name
    assert test_device._codename == _codename
    assert test_device.number == number
    assert test_device._protocol == protocol
    assert test_device.registers == registers

    assert bool(test_device)

    test_device.__del__()
    assert not bool(test_device)


@dataclass
class FakeReceiver:
    path: str = "11"
    handle: int = 0x11
    codename: Optional[str] = None

    def device_codename(self, number):
        return self.codename

    def __contains__(self, dev):
        return True


pi_CCCC = {"wpid": "CCCC", "kind": 0, "serial": None, "polling": "1ms", "power_switch": "top"}
pi_2011 = {"wpid": "2011", "kind": 1, "serial": "1234", "polling": "2ms", "power_switch": "bottom"}
pi_4066 = {"wpid": "4066", "kind": 1, "serial": "5678", "polling": "4ms", "power_switch": "left"}
pi_1007 = {"wpid": "1007", "kind": 2, "serial": "1234", "polling": "8ms", "power_switch": "right"}
pi_407B = {"wpid": "407B", "kind": 2, "serial": "5678", "polling": "1ms", "power_switch": "left"}
pi_DDDD = {"wpid": "DDDD", "kind": 2, "serial": "1234", "polling": "2ms", "power_switch": "top"}


@pytest.mark.parametrize(
    "number, pairing_info, responses, handle, _name, codename, p, p2, name",
    zip(
        range(1, 7),
        [pi_CCCC, pi_2011, pi_4066, pi_1007, pi_407B, pi_DDDD],
        [
            fake_hidpp.r_empty,
            fake_hidpp.r_keyboard_1,
            fake_hidpp.r_keyboard_2,
            fake_hidpp.r_mouse_1,
            fake_hidpp.r_mouse_2,
            fake_hidpp.r_mouse_3,
        ],
        [0x11, 0x11, 0x11, 0x11, 0x11, 0x11],
        [None, "Wireless Keyboard K520", "Craft Advanced Keyboard", "MX Air", "MX Vertical Wireless Mouse", None],
        ["CODE", "K520", "Craft", "MX Air", "MX Vertical", "CODE"],
        [None, 1.0, 4.5, 1.0, 4.5, None],
        [1.0, 1.0, 4.5, 1.0, 4.5, 4.5],
        [
            "CODE",
            "Wireless Keyboard K520",
            "Craft Advanced Keyboard",
            "MX Air",
            "MX Vertical Wireless Mouse",
            "ABABABABABABABADED",
        ],
    ),
)
def test_device_receiver(number, pairing_info, responses, handle, _name, codename, p, p2, name):
    low_level = LowLevelInterfaceFake(responses)
    low_level.request = partial(fake_hidpp.request, fake_hidpp.replace_number(responses, number))
    low_level.ping = partial(fake_hidpp.ping, fake_hidpp.replace_number(responses, number))

    test_device = device.Device(low_level, FakeReceiver(codename="CODE"), number, True, pairing_info, handle=handle)
    test_device.receiver.device = test_device

    assert test_device.handle == handle
    assert test_device._name == _name
    assert test_device.codename == codename
    assert test_device.number == number
    assert test_device._protocol == p
    assert test_device.protocol == p2
    assert test_device.codename == codename
    assert test_device.name == name

    assert test_device == test_device
    assert not (test_device != test_device)
    assert bool(test_device)

    test_device.__del__()


@pytest.mark.parametrize(
    "number, info, responses, handle, unitId, modelId, task_id, kind, firmware, serial, id, psl, rate",
    zip(
        range(1, 7),
        [pi_CCCC, pi_2011, pi_4066, pi_1007, pi_407B, pi_DDDD],
        [
            fake_hidpp.r_empty,
            fake_hidpp.r_keyboard_1,
            fake_hidpp.r_keyboard_2,
            fake_hidpp.r_mouse_1,
            fake_hidpp.r_mouse_2,
            fake_hidpp.r_mouse_3,
        ],
        [None, 0x11, 0x11, 0x11, 0x11, 0x11],
        [None, None, "12345678", None, None, "12345679"],  # unitId
        [None, None, "1234567890AB", None, None, "123456780000"],  # modelId
        [None, None, {"btid": "1234", "wpid": "5678", "usbid": "90AB"}, None, None, {"usbid": "1234"}],  # tid_map
        ["?", 1, 1, 2, 2, 2],  # kind
        [(), True, (), (), (), True],  # firmware
        [None, "1234", "5678", "1234", "5678", "1234"],  # serial
        ["", "1234", "12345678", "1234", "5678", "12345679"],  # id
        ["top", "bottom", "left", "right", "left", "top"],  # power switch location
        ["1ms", "2ms", "4ms", "8ms", "1ms", "9ms"],  # polling rate
    ),
)
def test_device_ids(number, info, responses, handle, unitId, modelId, task_id, kind, firmware, serial, id, psl, rate):
    low_level = LowLevelInterfaceFake(responses)
    low_level.request = partial(fake_hidpp.request, fake_hidpp.replace_number(responses, number))
    low_level.ping = partial(fake_hidpp.ping, fake_hidpp.replace_number(responses, number))

    test_device = device.Device(low_level, FakeReceiver(), number, True, info, handle=handle)

    assert test_device.unitId == unitId
    assert test_device.modelId == modelId
    assert test_device.tid_map == task_id
    assert test_device.kind == kind
    assert test_device.firmware == firmware or len(test_device.firmware) > 0 and firmware is True
    assert test_device.id == id
    assert test_device.power_switch_location == psl
    assert test_device.polling_rate == rate


class FakeDevice(device.Device):  # a fully functional Device but its HID++ functions look at local data
    def __init__(self, responses, *args, **kwargs):
        self.responses = responses
        super().__init__(LowLevelInterfaceFake(responses), *args, **kwargs)

    request = fake_hidpp.Device.request
    ping = fake_hidpp.Device.ping


@pytest.mark.parametrize(
    "device_info, responses, protocol, led, keys, remap, gestures, backlight, profiles",
    [
        (di_CCCC, fake_hidpp.r_empty, 1.0, type(None), None, None, None, None, None),
        (di_C318, fake_hidpp.r_empty, 1.0, type(None), None, None, None, None, None),
        (di_B530, fake_hidpp.r_keyboard_1, 1.0, type(None), None, None, None, None, None),
        (di_B530, fake_hidpp.r_keyboard_2, 2.0, type(None), 4, 0, 0, None, None),
        (di_B530, fake_hidpp.complex_responses_1, 4.5, hidpp20.LEDEffectsInfo, 0, 0, 0, None, None),
        (di_B530, fake_hidpp.complex_responses_2, 4.5, hidpp20.RGBEffectsInfo, 8, 3, 1, True, True),
    ],
)
def test_device_complex(device_info, responses, protocol, led, keys, remap, gestures, backlight, profiles, mocker):
    test_device = FakeDevice(responses, None, None, True, device_info=device_info)
    test_device._name = "TestDevice"
    test_device._protocol = protocol
    spy_request = mocker.spy(test_device, "request")

    assert type(test_device.led_effects) == led
    if keys is None:
        assert test_device.keys == keys
    else:
        assert len(test_device.keys) == keys
    if remap is None:
        assert test_device.remap_keys == remap
    else:
        assert len(test_device.remap_keys) == remap
    assert (test_device.gestures is None) == (gestures is None)
    assert (test_device.backlight is None) == (backlight is None)
    assert (test_device.profiles is None) == (profiles is None)

    test_device.set_configuration(55)
    if protocol > 1.0:
        spy_request.assert_called_with(0x210, 55, no_reply=False)
    test_device.reset()
    if protocol > 1.0:
        spy_request.assert_called_with(0x210, 0, no_reply=False)


@pytest.mark.parametrize(
    "device_info, responses, protocol, p, persister, settings",
    [
        (di_CCCC, fake_hidpp.r_empty, 1.0, None, None, 0),
        (di_C318, fake_hidpp.r_empty, 1.0, {}, {}, 0),
        (di_C318, fake_hidpp.r_keyboard_1, 1.0, {"n": "n"}, {"n": "n"}, 1),
        (di_B530, fake_hidpp.r_keyboard_2, 4.5, {"m": "m"}, {"m": "m"}, 1),
        (di_C068, fake_hidpp.r_mouse_1, 1.0, {"o": "o"}, {"o": "o"}, 2),
        (di_C08A, fake_hidpp.r_mouse_2, 4.5, {"p": "p"}, {"p": "p"}, 0),
    ],
)
def test_device_settings(device_info, responses, protocol, p, persister, settings, mocker):
    mocker.patch("solaar.configuration.persister", return_value=p)
    test_device = FakeDevice(responses, None, None, True, device_info=device_info)
    test_device._name = "TestDevice"
    test_device._protocol = protocol

    assert test_device.persister == persister
    assert len(test_device.settings) == settings


@pytest.mark.parametrize(
    "device_info, responses, protocol, expected_battery, changed",
    [
        (di_C318, fake_hidpp.r_empty, 1.0, None, {"active": True, "alert": 0, "reason": None}),
        (
            di_C318,
            fake_hidpp.r_keyboard_1,
            1.0,
            common.Battery(BatteryLevelApproximation.GOOD.value, None, BatteryStatus.DISCHARGING, None),
            {"active": True, "alert": 0, "reason": None},
        ),
        (
            di_B530,
            fake_hidpp.r_keyboard_2,
            4.5,
            common.Battery(18, 52, None, None),
            {"active": True, "alert": 0, "reason": None},
        ),
    ],
)
def test_device_battery(device_info, responses, protocol, expected_battery, changed, mocker):
    test_device = FakeDevice(responses, None, None, online=True, device_info=device_info)
    test_device._name = "TestDevice"
    test_device._protocol = protocol
    spy_changed = mocker.spy(test_device, "changed")

    assert test_device.battery() == expected_battery
    test_device.read_battery()
    spy_changed.assert_called_with(**changed)



from dataclasses import dataclass
from dataclasses import field
from typing import Any
from typing import List
from typing import Optional

import pytest

from logitech_receiver import common
from logitech_receiver import hidpp10
from logitech_receiver import hidpp10_constants
from logitech_receiver.hidpp10_constants import PairingError
from logitech_receiver.hidpp10_constants import Registers

_hidpp10 = hidpp10.Hidpp10()


@dataclass
class Response:
    response: Optional[str]
    request_id: int
    params: Any


@dataclass
class Device:
    name: str = "Device"
    online: bool = True
    kind: str = "fake"
    protocol: float = 1.0
    isDevice: bool = False  # incorrect, but useful here
    registers: List[Registers] = field(default_factory=list)
    responses: List[Response] = field(default_factory=list)

    def request(self, id, params=None, no_reply=False):
        if params is None:
            params = []
        print("REQUEST ", self.name, hex(id), params)
        for r in self.responses:
            if id == r.request_id and params == r.params:
                print("RESPONSE", self.name, hex(r.request_id), r.params, r.response)
                return bytes.fromhex(r.response) if r.response is not None else None


device_offline = Device("OFFLINE", False)
device_leds = Device("LEDS", True, registers=[Registers.THREE_LEDS, Registers.BATTERY_STATUS])
device_features = Device("FEATURES", True, protocol=4.5)

registers_standard = [Registers.BATTERY_STATUS, Registers.FIRMWARE]
responses_standard = [
    Response("555555", 0x8100 | Registers.BATTERY_STATUS, 0x00),
    Response("666666", 0x8100 | Registers.BATTERY_STATUS, 0x10),
    Response("777777", 0x8000 | Registers.BATTERY_STATUS, 0x00),
    Response("888888", 0x8000 | Registers.BATTERY_STATUS, 0x10),
    Response("052100", 0x8100 | Registers.BATTERY_STATUS, []),
    Response("ABCDEF", 0x8100 | Registers.FIRMWARE, 0x01),
    Response("ABCDEF", 0x8100 | Registers.FIRMWARE, 0x02),
    Response("ABCDEF", 0x8100 | Registers.FIRMWARE, 0x03),
    Response("ABCDEF", 0x8100 | Registers.FIRMWARE, 0x04),
    Response("000900", 0x8100 | Registers.NOTIFICATIONS, []),
    Response("101010", 0x8100 | Registers.MOUSE_BUTTON_FLAGS, []),
    Response("010101", 0x8100 | Registers.KEYBOARD_FN_SWAP, []),
    Response("020202", 0x8100 | Registers.DEVICES_CONFIGURATION, []),
    Response("030303", 0x8000 | Registers.DEVICES_CONFIGURATION, 0x00),
]
device_standard = Device("STANDARD", True, registers=registers_standard, responses=responses_standard)


@pytest.mark.parametrize(
    "device, register, param, expected_result",
    [
        (device_offline, Registers.THREE_LEDS, 0x00, None),
        (device_standard, Registers.THREE_LEDS, 0x00, None),
        (device_standard, Registers.BATTERY_STATUS, 0x00, "555555"),
        (device_standard, Registers.BATTERY_STATUS, 0x10, "666666"),
    ],
)
def test_read_register(device, register, param, expected_result, mocker):
    spy_request = mocker.spy(device, "request")

    result = hidpp10.read_register(device, register, param)

    assert result == (bytes.fromhex(expected_result) if expected_result else None)
    spy_request.assert_called_once_with(0x8100 | register, param)


@pytest.mark.parametrize(
    "device, register, param, expected_result",
    [
        (device_offline, Registers.THREE_LEDS, 0x00, None),
        (device_standard, Registers.THREE_LEDS, 0x00, None),
        (device_standard, Registers.BATTERY_STATUS, 0x00, "777777"),
        (device_standard, Registers.BATTERY_STATUS, 0x10, "888888"),
    ],
)
def test_write_register(device, register, param, expected_result, mocker):
    spy_request = mocker.spy(device, "request")

    result = hidpp10.write_register(device, register, param)

    assert result == (bytes.fromhex(expected_result) if expected_result else None)
    spy_request.assert_called_once_with(0x8000 | register, param)


def device_charge(name, response):
    responses = [Response(response, 0x8100 | Registers.BATTERY_CHARGE, [])]
    return Device(name, registers=[], responses=responses)


device_charge1 = device_charge("DISCHARGING", "550030")
device_charge2 = device_charge("RECHARGING", "440050")
device_charge3 = device_charge("FULL", "600090")
device_charge4 = device_charge("OTHER", "220000")


def device_status(name, response):
    responses = [Response(response, 0x8100 | Registers.BATTERY_STATUS, [])]
    return Device(name, registers=[], responses=responses)


device_status1 = device_status("FULL", "072200")
device_status2 = device_status("GOOD", "052100")
device_status3 = device_status("LOW", "032200")
device_status4 = device_status("CRITICAL", "010100")
device_status5 = device_status("EMPTY", "000000")
device_status6 = device_status("NOSTATUS", "002200")


@pytest.mark.parametrize(
    "device, expected_result, expected_register",
    [
        (device_offline, None, None),
        (device_features, None, None),
        (device_leds, None, None),
        (
            device_standard,
            common.Battery(common.BatteryLevelApproximation.GOOD, None, common.BatteryStatus.RECHARGING, None),
            Registers.BATTERY_STATUS,
        ),
        (device_charge1, common.Battery(0x55, None, common.BatteryStatus.DISCHARGING, None), Registers.BATTERY_CHARGE),
        (
            device_charge2,
            common.Battery(0x44, None, common.BatteryStatus.RECHARGING, None),
            Registers.BATTERY_CHARGE,
        ),
        (
            device_charge3,
            common.Battery(0x60, None, common.BatteryStatus.FULL, None),
            Registers.BATTERY_CHARGE,
        ),
        (device_charge4, common.Battery(0x22, None, None, None), Registers.BATTERY_CHARGE),
        (
            device_status1,
            common.Battery(common.BatteryLevelApproximation.FULL, None, common.BatteryStatus.FULL, None),
            Registers.BATTERY_STATUS,
        ),
        (
            device_status2,
            common.Battery(common.BatteryLevelApproximation.GOOD, None, common.BatteryStatus.RECHARGING, None),
            Registers.BATTERY_STATUS,
        ),
        (
            device_status3,
            common.Battery(common.BatteryLevelApproximation.LOW, None, common.BatteryStatus.FULL, None),
            Registers.BATTERY_STATUS,
        ),
        (
            device_status4,
            common.Battery(common.BatteryLevelApproximation.CRITICAL, None, None, None),
            Registers.BATTERY_STATUS,
        ),
        (
            device_status5,
            common.Battery(common.BatteryLevelApproximation.EMPTY, None, common.BatteryStatus.DISCHARGING, None),
            Registers.BATTERY_STATUS,
        ),
        (
            device_status6,
            common.Battery(None, None, common.BatteryStatus.FULL, None),
            Registers.BATTERY_STATUS,
        ),
    ],
)
def test_hidpp10_get_battery(device, expected_result, expected_register):
    result = _hidpp10.get_battery(device)

    assert result == expected_result
    if expected_register is not None:
        assert expected_register in device.registers


@pytest.mark.parametrize(
    "device, expected_firmwares",
    [
        (device_offline, []),
        (
            device_standard,
            [
                common.FirmwareKind.Firmware,
                common.FirmwareKind.Bootloader,
                common.FirmwareKind.Other,
            ],
        ),
    ],
)
def test_hidpp10_get_firmware(device, expected_firmwares):
    firmwares = _hidpp10.get_firmware(device)

    if not expected_firmwares:
        assert firmwares is None
    else:
        firmware_types = [firmware.kind for firmware in firmwares]
        assert firmware_types == expected_firmwares
        assert len(firmwares) == len(expected_firmwares)


@pytest.mark.parametrize(
    "device, level, charging, warning, p1, p2",
    [
        (device_leds, common.BatteryLevelApproximation.EMPTY, False, False, 0x33, 0x00),
        (device_leds, common.BatteryLevelApproximation.CRITICAL, False, False, 0x22, 0x00),
        (device_leds, common.BatteryLevelApproximation.LOW, False, False, 0x20, 0x00),
        (device_leds, common.BatteryLevelApproximation.GOOD, False, False, 0x20, 0x02),
        (device_leds, common.BatteryLevelApproximation.FULL, False, False, 0x20, 0x22),
        (device_leds, None, True, False, 0x30, 0x33),
        (device_leds, None, False, True, 0x02, 0x00),
        (device_leds, None, False, False, 0x11, 0x11),
    ],
)
def test_set_3leds(device, level, charging, warning, p1, p2, mocker):
    spy_request = mocker.spy(device, "request")

    _hidpp10.set_3leds(device, level, charging, warning)

    spy_request.assert_called_once_with(0x8000 | Registers.THREE_LEDS, p1, p2)


@pytest.mark.parametrize("device", [device_offline, device_features])
def test_set_3leds_missing(device, mocker):
    spy_request = mocker.spy(device, "request")

    _hidpp10.set_3leds(device)

    assert spy_request.call_count == 0


@pytest.mark.parametrize("device", [device_standard])
def test_get_notification_flags(device):
    result = _hidpp10.get_notification_flags(device)

    assert result == hidpp10_constants.NotificationFlag(int("000900", 16))


def test_set_notification_flags(mocker):
    device = device_standard
    spy_request = mocker.spy(device, "request")

    result = _hidpp10.set_notification_flags(
        device, hidpp10_constants.NotificationFlag.BATTERY_STATUS, hidpp10_constants.NotificationFlag.WIRELESS
    )

    spy_request.assert_called_once_with(0x8000 | Registers.NOTIFICATIONS, b"\x10\x01\x00")
    assert result is not None


def test_set_notification_flags_bad(mocker):
    device = device_features
    spy_request = mocker.spy(device, "request")

    result = _hidpp10.set_notification_flags(
        device, hidpp10_constants.NotificationFlag.BATTERY_STATUS, hidpp10_constants.NotificationFlag.WIRELESS
    )

    assert spy_request.call_count == 0
    assert result is None


@pytest.mark.parametrize(
    "flag_bits, expected_names",
    [
        # doesn't work in Python 3.8 and 3.10 for some reason        (None, ""),
        (hidpp10_constants.NotificationFlag(0x0), "none"),
        (hidpp10_constants.NotificationFlag(0x001000), "multi touch"),
        (hidpp10_constants.NotificationFlag(0x080000), "mouse extra buttons"),
        (
            hidpp10_constants.NotificationFlag(0x080400),
            ("link quality\n               mouse extra buttons"),
        ),
    ],
)
def test_notification_flag_str(flag_bits, expected_names):
    flag_names = hidpp10_constants.flags_to_str(
        hidpp10_constants.NotificationFlag(flag_bits) if flag_bits is not None else None, fallback="none"
    )

    assert flag_names == expected_names


def test_get_device_features():
    result = _hidpp10.get_device_features(device_standard)

    assert result == int("101010", 16)


@pytest.mark.parametrize(
    "device, register, expected_result",
    [
        (device_standard, Registers.BATTERY_STATUS, "052100"),
        (device_standard, Registers.MOUSE_BUTTON_FLAGS, "101010"),
        (device_standard, Registers.KEYBOARD_ILLUMINATION, None),
        (device_features, Registers.KEYBOARD_ILLUMINATION, None),
    ],
)
def test_get_register(device, register, expected_result):
    result = _hidpp10._get_register(device, register)

    assert result == (int(expected_result, 16) if expected_result is not None else None)


@pytest.mark.parametrize(
    "device, expected_result",
    [
        (device_standard, 2),
        (device_features, None),
    ],
)
def test_get_configuration_pending_flags(device, expected_result):
    result = hidpp10.get_configuration_pending_flags(device)

    assert result == expected_result


@pytest.mark.parametrize(
    "device, expected_result",
    [
        (device_standard, True),
        (device_features, False),
    ],
)
def test_set_configuration_pending_flags(device, expected_result):
    result = hidpp10.set_configuration_pending_flags(device, 0x00)

    assert result == expected_result


def test_pairing_error():
    expected_label = "device not supported"

    res = PairingError.DEVICE_NOT_SUPPORTED.label

    assert res == expected_label



from enum import IntFlag

import pytest
import yaml

from logitech_receiver import common


def test_crc16():
    value = b"123456789"
    expected = 0x29B1

    result = common.crc16(value)

    assert result == expected


def test_named_int():
    named_int = common.NamedInt(0x2, "pulse")

    assert named_int.name == "pulse"
    assert named_int == 2
    assert repr(named_int) == "NamedInt(2, 'pulse')"


def test_named_int_comparison():
    named_int = common.NamedInt(0, "entry")
    named_int_equal = common.NamedInt(0, "entry")
    named_int_unequal_name = common.NamedInt(0, "unequal")
    named_int_unequal_value = common.NamedInt(5, "entry")
    named_int_unequal = common.NamedInt(2, "unequal")

    assert named_int == named_int_equal
    assert named_int != named_int_unequal_name
    assert named_int != named_int_unequal_value
    assert named_int != named_int_unequal
    assert named_int is not None
    assert named_int == 0
    assert named_int == "entry"


def test_named_int_comparison_exception():
    named_int = common.NamedInt(0, "entry")

    with pytest.raises(TypeError):
        assert named_int == b"\x00"


def test_named_int_conversions():
    named_int = common.NamedInt(2, "two")

    assert named_int.bytes() == b"\x00\x02"
    assert str(named_int) == "two"


def test_named_int_yaml():
    named_int = common.NamedInt(2, "two")

    yaml_string = yaml.dump(named_int)

    #    assert yaml_string == "!NamedInt {name: two, value: 2}\n"

    yaml_load = yaml.safe_load(yaml_string)

    assert yaml_load == named_int


def test_named_ints():
    named_ints = common.NamedInts(empty=0, critical=5, low=20, good=50, full=90)

    assert named_ints.empty == 0
    assert named_ints.empty.name == "empty"
    assert named_ints.critical == 5
    assert named_ints.critical.name == "critical"
    assert named_ints.low == 20
    assert named_ints.low.name == "low"
    assert named_ints.good == 50
    assert named_ints.good.name == "good"
    assert named_ints.full == 90
    assert named_ints.full.name == "full"

    assert len(named_ints) == 5
    assert 5 in named_ints
    assert 6 not in named_ints
    assert "critical" in named_ints
    assert "easy" not in named_ints
    assert common.NamedInt(5, "critical") in named_ints
    assert common.NamedInt(5, "five") not in named_ints
    assert common.NamedInt(6, "critical") not in named_ints
    assert named_ints[5] == "critical"
    assert named_ints["critical"] == "critical"
    assert named_ints[66] is None
    assert named_ints["5"] is None
    assert len(named_ints[:]) == len(named_ints)
    assert len(named_ints[0:100]) == len(named_ints)
    assert len(named_ints[5:90]) == 3
    assert len(named_ints[5:]) == 4
    assert named_ints[90:5] == []


def test_named_ints_fallback():
    named_ints = common.NamedInts(empty=0, critical=5, low=20, good=50, full=90)
    named_ints._fallback = lambda x: str(x)

    fallback = named_ints[80]

    assert fallback == common.NamedInt(80, "80")


def test_named_ints_list():
    named_ints_list = common.NamedInts.list([0, 5, 20, 50, 90])

    assert len(named_ints_list) == 5
    assert 50 in named_ints_list
    assert 60 not in named_ints_list


def test_named_ints_range():
    named_ints_range = common.NamedInts.range(0, 5)

    assert len(named_ints_range) == 6
    assert 4 in named_ints_range
    assert 6 not in named_ints_range


@pytest.mark.parametrize(
    "code, expected_flags",
    [
        (0, []),
        (0b0010, ["two"]),
        (0b0101, ["one", "three"]),
        (0b1001, ["one", "unknown:000008"]),
    ],
)
def test_named_ints_flag_names(code, expected_flags):
    named_ints_flag_bits = common.NamedInts(one=0b001, two=0b010, three=0b100)

    flags = list(named_ints_flag_bits.flag_names(code))

    assert flags == expected_flags


@pytest.mark.parametrize(
    "code, expected_flags",
    [
        (0, []),
        (0b0010, ["two"]),
        (0b0101, ["one", "three"]),
        (0b1001, ["one", "unknown:000008"]),
    ],
)
def test_flag_names(code, expected_flags):
    class ExampleFlag(IntFlag):
        one = 0x1
        two = 0x2
        three = 0x4

    flags = common.flag_names(ExampleFlag, code)

    assert list(flags) == expected_flags


def test_named_ints_setitem():
    named_ints = common.NamedInts(empty=0, critical=5, low=20, good=50, full=90)

    named_ints[55] = "better"
    named_ints[60] = common.NamedInt(60, "sixty")
    with pytest.raises(TypeError):
        named_ints[70] = 70
    with pytest.raises(ValueError):
        named_ints[70] = "empty"
    with pytest.raises(ValueError):
        named_ints[50] = "new"

    assert named_ints[55] == "better"
    assert named_ints[60] == "sixty"


def test_named_ints_other():
    named_ints = common.NamedInts(empty=0, critical=5)
    named_ints_2 = common.NamedInts(good=50)

    union = named_ints.__or__(named_ints_2)

    assert list(named_ints) == [common.NamedInt(0, "empty"), common.NamedInt(5, "critical")]
    assert len(named_ints) == 2
    assert repr(named_ints) == "NamedInts(NamedInt(0, 'empty'), NamedInt(5, 'critical'))"
    assert len(union) == 3
    assert list(union) == [common.NamedInt(0, "empty"), common.NamedInt(5, "critical"), common.NamedInt(50, "good")]


def test_unsorted_named_ints():
    named_ints = common.UnsortedNamedInts(critical=5, empty=0)
    named_ints_2 = common.UnsortedNamedInts(good=50)

    union = named_ints.__or__(named_ints_2)
    unionr = named_ints_2.__or__(named_ints)

    assert len(union) == 3
    assert list(union) == [common.NamedInt(5, "critical"), common.NamedInt(0, "empty"), common.NamedInt(50, "good")]
    assert len(unionr) == 3
    assert list(unionr) == [common.NamedInt(50, "good"), common.NamedInt(5, "critical"), common.NamedInt(0, "empty")]


@pytest.mark.parametrize(
    "bytes_input, expected_output",
    [
        (b"\x01\x02\x03\x04", "01020304"),
        (b"", ""),
    ],
)
def test_strhex(bytes_input, expected_output):
    result = common.strhex(bytes_input)

    assert result == expected_output


def test_bytest2int():
    value = b"\x12\x34\x56\x78"
    expected = 0x12345678

    result = common.bytes2int(value)

    assert result == expected


def test_int2bytes():
    value = 0x12345678
    expected = b"\x12\x34\x56\x78"

    result = common.int2bytes(value)

    assert result == expected


def test_kw_exception():
    e = common.KwException(foo=0, bar="bar")

    assert e.foo == 0
    assert e.bar == "bar"


@pytest.mark.parametrize(
    "status, expected_level, expected_ok, expected_charging, expected_string",
    [
        (common.BatteryStatus.FULL, common.BatteryLevelApproximation.FULL, True, True, "Battery: full (full)"),
        (common.BatteryStatus.ALMOST_FULL, common.BatteryLevelApproximation.GOOD, True, True, "Battery: good (almost full)"),
        (common.BatteryStatus.RECHARGING, common.BatteryLevelApproximation.GOOD, True, True, "Battery: good (recharging)"),
        (
            common.BatteryStatus.SLOW_RECHARGE,
            common.BatteryLevelApproximation.LOW,
            True,
            True,
            "Battery: low (slow recharge)",
        ),
        (common.BatteryStatus.DISCHARGING, None, True, False, ""),
    ],
)
def test_battery(status, expected_level, expected_ok, expected_charging, expected_string):
    battery = common.Battery(None, None, status, None)

    assert battery.status == status
    assert battery.level == expected_level
    assert battery.ok() == expected_ok
    assert battery.charging() == expected_charging
    assert battery.to_str() == expected_string


def test_battery_2():
    battery = common.Battery(50, None, common.BatteryStatus.DISCHARGING, None)

    assert battery.status == common.BatteryStatus.DISCHARGING
    assert battery.level == 50
    assert battery.ok()
    assert not battery.charging()
    assert battery.to_str() == "Battery: 50% (discharging)"



from unittest import mock

from logitech_receiver import desktop_notifications

# depends on external environment, so make some tests dependent on availability


def test_init():
    result = desktop_notifications.init()

    assert result == desktop_notifications.available


def test_uninit():
    assert desktop_notifications.uninit() is None


class MockDevice(mock.Mock):
    name = "MockDevice"

    def close():
        return True


def test_show():
    dev = MockDevice()
    reason = "unknown"
    result = desktop_notifications.show(dev, reason)
    assert result is not None if desktop_notifications.available else result is None



import locale
import os
import platform

import pytest

from solaar import i18n


@pytest.fixture
def set_locale_de():
    backup_lang = os.environ.get("LC_ALL", "")
    try:
        yield
    finally:
        os.environ["LC_ALL"] = backup_lang
        i18n.set_locale_to_system_default()


@pytest.mark.skipif(platform.system() == "Linux", reason="Adapt test for Linux")
def test_set_locale_to_system_default(set_locale_de):
    os.environ["LC_ALL"] = "de_DE.UTF-8"
    i18n.set_locale_to_system_default()

    language, encoding = locale.getlocale()

    assert language == "de_DE"
    assert encoding == "UTF-8"



import pytest

from logitech_receiver import notifications
from logitech_receiver.base import HIDPPNotification
from logitech_receiver.common import Notification
from logitech_receiver.hidpp10_constants import BoltPairingError
from logitech_receiver.hidpp10_constants import PairingError
from logitech_receiver.hidpp10_constants import Registers
from logitech_receiver.hidpp20_constants import SupportedFeature
from logitech_receiver.receiver import Receiver

from . import fake_hidpp


class MockLowLevelInterface:
    def open_path(self, path):
        pass

    def find_paired_node_wpid(self, receiver_path: str, index: int):
        pass

    def ping(self, handle, number, long_message=False):
        pass

    def request(self, handle, devnumber, request_id, *params, **kwargs):
        pass

    def find_paired_node(self, receiver_path: str, index: int, timeout: int):
        return None

    def close(self, device_handle) -> None:
        pass


@pytest.mark.parametrize(
    "sub_id, notification_data, expected_error, expected_new_device",
    [
        (Registers.DISCOVERY_STATUS_NOTIFICATION, b"\x01", BoltPairingError.DEVICE_TIMEOUT, None),
        (
            Registers.DEVICE_DISCOVERY_NOTIFICATION,
            b"\x01\x01\x01\x01\x01\x01\x01\x01\x01",
            None,
            None,
        ),
        (Registers.PAIRING_STATUS_NOTIFICATION, b"\x02", BoltPairingError.FAILED, None),
        (Notification.PAIRING_LOCK, b"\x01", PairingError.DEVICE_TIMEOUT, None),
        (Notification.PAIRING_LOCK, b"\x02", PairingError.DEVICE_NOT_SUPPORTED, None),
        (Notification.PAIRING_LOCK, b"\x03", PairingError.TOO_MANY_DEVICES, None),
        (Notification.PAIRING_LOCK, b"\x06", PairingError.SEQUENCE_TIMEOUT, None),
        (Registers.PASSKEY_REQUEST_NOTIFICATION, b"\x06", None, None),
        (Registers.PASSKEY_PRESSED_NOTIFICATION, b"\x06", None, None),
    ],
)
def test_process_receiver_notification(sub_id, notification_data, expected_error, expected_new_device):
    receiver: Receiver = Receiver(MockLowLevelInterface(), None, {}, True, None, None)
    notification = HIDPPNotification(0, 0, sub_id, 0x02, notification_data)

    result = notifications.process_receiver_notification(receiver, notification)

    assert result
    assert receiver.pairing.error == (None if expected_error is None else expected_error.label)
    assert receiver.pairing.new_device is expected_new_device


@pytest.mark.parametrize(
    "hidpp_notification, expected",
    [
        (HIDPPNotification(0, 0, sub_id=Registers.BATTERY_STATUS, address=0, data=b"0x01"), False),
        (HIDPPNotification(0, 0, sub_id=Notification.NO_OPERATION, address=0, data=b"0x01"), False),
        (HIDPPNotification(0, 0, sub_id=0x40, address=0, data=b"0x01"), True),
    ],
)
def test_process_device_notification(hidpp_notification, expected):
    device = fake_hidpp.Device()

    result = notifications.process_device_notification(device, hidpp_notification)

    assert result == expected


@pytest.mark.parametrize(
    "hidpp_notification, expected",
    [
        (HIDPPNotification(0, 0, sub_id=Notification.CONNECT_DISCONNECT, address=0, data=b"0x01"), True),
        (HIDPPNotification(0, 0, sub_id=Notification.DJ_PAIRING, address=0, data=b"0x01"), True),
        (HIDPPNotification(0, 0, sub_id=Notification.CONNECTED, address=0, data=b"0x01"), True),
        (HIDPPNotification(0, 0, sub_id=Notification.RAW_INPUT, address=0, data=b"0x01"), None),
    ],
)
def test_process_dj_notification(hidpp_notification, expected):
    device = fake_hidpp.Device()

    result = notifications._process_dj_notification(device, hidpp_notification)

    assert result == expected


@pytest.mark.parametrize(
    "hidpp_notification, expected",
    [
        (HIDPPNotification(0, 0, sub_id=Registers.BATTERY_STATUS, address=0, data=b"\x01\x00"), True),
        (HIDPPNotification(0, 0, sub_id=Registers.BATTERY_CHARGE, address=0, data=b"0x01\x00"), True),
        (HIDPPNotification(0, 0, sub_id=Notification.RAW_INPUT, address=0, data=b"0x01"), None),
    ],
)
def test_process_hidpp10_custom_notification(hidpp_notification, expected):
    device = fake_hidpp.Device()

    result = notifications._process_hidpp10_custom_notification(device, hidpp_notification)

    assert result == expected


@pytest.mark.parametrize(
    "hidpp_notification, expected",
    [
        (HIDPPNotification(0, 0, sub_id=Notification.CONNECT_DISCONNECT, address=0x02, data=b"0x01"), True),
        (HIDPPNotification(0, 0, sub_id=Notification.CONNECT_DISCONNECT, address=0x00, data=b"0x01"), True),
        (HIDPPNotification(0, 0, sub_id=Notification.DJ_PAIRING, address=0x00, data=b"0x01"), True),
        (HIDPPNotification(0, 0, sub_id=Notification.DJ_PAIRING, address=0x02, data=b"0x01"), True),
        (HIDPPNotification(0, 0, sub_id=Notification.DJ_PAIRING, address=0x03, data=b"0x01"), True),
        (HIDPPNotification(0, 0, sub_id=Notification.DJ_PAIRING, address=0x03, data=b"0x4040"), True),
        (HIDPPNotification(0, 0, sub_id=Notification.RAW_INPUT, address=0x00, data=b"0x01"), True),
        (HIDPPNotification(0, 0, sub_id=Notification.POWER, address=0x00, data=b"0x01"), True),
        (HIDPPNotification(0, 0, sub_id=Notification.POWER, address=0x01, data=b"0x01"), True),
        (HIDPPNotification(0, 0, sub_id=Notification.PAIRING_LOCK, address=0x01, data=b"0x01"), None),
    ],
)
def test_process_hidpp10_notification(hidpp_notification, expected):
    fake_device = fake_hidpp.Device()
    fake_device.receiver = ["rec1", "rec2"]

    result = notifications._process_hidpp10_notification(fake_device, hidpp_notification)

    assert result == expected


@pytest.mark.parametrize(
    "hidpp_notification, feature",
    [
        (
            HIDPPNotification(0, 0, sub_id=Notification.CONNECT_DISCONNECT, address=0x00, data=b"0x01"),
            SupportedFeature.BATTERY_STATUS,
        ),
        (
            HIDPPNotification(0, 0, sub_id=Notification.CONNECT_DISCONNECT, address=0x02, data=b"0x01"),
            SupportedFeature.BATTERY_STATUS,
        ),
        (
            HIDPPNotification(0, 0, sub_id=Notification.CONNECT_DISCONNECT, address=0x00, data=b"0x01"),
            SupportedFeature.BATTERY_VOLTAGE,
        ),
        (
            HIDPPNotification(0, 0, sub_id=Notification.CONNECT_DISCONNECT, address=0x05, data=b"0x01"),
            SupportedFeature.BATTERY_VOLTAGE,
        ),
        (
            HIDPPNotification(0, 0, sub_id=Notification.CONNECT_DISCONNECT, address=0x00, data=b"0x01"),
            SupportedFeature.UNIFIED_BATTERY,
        ),
        (
            HIDPPNotification(0, 0, sub_id=Notification.CONNECT_DISCONNECT, address=0x02, data=b"0x01"),
            SupportedFeature.UNIFIED_BATTERY,
        ),
        (
            HIDPPNotification(0, 0, sub_id=Notification.CONNECT_DISCONNECT, address=0x00, data=b"0x01"),
            SupportedFeature.ADC_MEASUREMENT,
        ),
        (
            HIDPPNotification(0, 0, sub_id=Notification.CONNECT_DISCONNECT, address=0x02, data=b"0x01"),
            SupportedFeature.ADC_MEASUREMENT,
        ),
        (
            HIDPPNotification(0, 0, sub_id=Notification.CONNECT_DISCONNECT, address=0x00, data=b"01234GOOD"),
            SupportedFeature.SOLAR_DASHBOARD,
        ),
        (
            HIDPPNotification(0, 0, sub_id=Notification.CONNECT_DISCONNECT, address=0x10, data=b"01234GOOD"),
            SupportedFeature.SOLAR_DASHBOARD,
        ),
        (
            HIDPPNotification(0, 0, sub_id=Notification.CONNECT_DISCONNECT, address=0x20, data=b"01234GOOD"),
            SupportedFeature.SOLAR_DASHBOARD,
        ),
        (
            HIDPPNotification(0, 0, sub_id=Notification.CONNECT_DISCONNECT, address=0x02, data=b"01234GOOD"),
            SupportedFeature.SOLAR_DASHBOARD,
        ),
        (
            HIDPPNotification(0, 0, sub_id=Notification.CONNECT_DISCONNECT, address=0x02, data=b"CHARGENOTGOOD"),
            SupportedFeature.SOLAR_DASHBOARD,
        ),
        (
            HIDPPNotification(0, 0, sub_id=Notification.CONNECT_DISCONNECT, address=0x00, data=b"\x01\x01\x02"),
            SupportedFeature.WIRELESS_DEVICE_STATUS,
        ),
        (
            HIDPPNotification(0, 0, sub_id=Notification.CONNECT_DISCONNECT, address=0x02, data=b"0x01"),
            SupportedFeature.WIRELESS_DEVICE_STATUS,
        ),
        (
            HIDPPNotification(0, 0, sub_id=Notification.CONNECT_DISCONNECT, address=0x00, data=b"0x01"),
            SupportedFeature.TOUCHMOUSE_RAW_POINTS,
        ),
        (
            HIDPPNotification(0, 0, sub_id=Notification.CONNECT_DISCONNECT, address=0x10, data=b"0x01"),
            SupportedFeature.TOUCHMOUSE_RAW_POINTS,
        ),
        (
            HIDPPNotification(0, 0, sub_id=Notification.CONNECT_DISCONNECT, address=0x05, data=b"0x01"),
            SupportedFeature.TOUCHMOUSE_RAW_POINTS,
        ),
        (
            HIDPPNotification(0, 0, sub_id=Notification.CONNECT_DISCONNECT, address=0x00, data=b"0x01"),
            SupportedFeature.REPROG_CONTROLS,
        ),
        (
            HIDPPNotification(0, 0, sub_id=Notification.CONNECT_DISCONNECT, address=0x02, data=b"0x01"),
            SupportedFeature.REPROG_CONTROLS,
        ),
        (
            HIDPPNotification(0, 0, sub_id=Notification.CONNECT_DISCONNECT, address=0x00, data=b"0x01"),
            SupportedFeature.BACKLIGHT2,
        ),
        (
            HIDPPNotification(
                0, 0, sub_id=Notification.CONNECT_DISCONNECT, address=0x00, data=b"\x01\x01\x01\x01\x01\x01\x01\x01"
            ),
            SupportedFeature.REPROG_CONTROLS_V4,
        ),
        (
            HIDPPNotification(0, 0, sub_id=Notification.CONNECT_DISCONNECT, address=0x10, data=b"0x01"),
            SupportedFeature.REPROG_CONTROLS_V4,
        ),
        (
            HIDPPNotification(0, 0, sub_id=Notification.CONNECT_DISCONNECT, address=0x20, data=b"0x01"),
            SupportedFeature.REPROG_CONTROLS_V4,
        ),
        (
            HIDPPNotification(0, 0, sub_id=Notification.CONNECT_DISCONNECT, address=0x00, data=b"0x01"),
            SupportedFeature.HIRES_WHEEL,
        ),
        (
            HIDPPNotification(0, 0, sub_id=Notification.CONNECT_DISCONNECT, address=0x10, data=b"0x01"),
            SupportedFeature.HIRES_WHEEL,
        ),
        (
            HIDPPNotification(0, 0, sub_id=Notification.CONNECT_DISCONNECT, address=0x02, data=b"0x01"),
            SupportedFeature.HIRES_WHEEL,
        ),
        (
            HIDPPNotification(0, 0, sub_id=Notification.CONNECT_DISCONNECT, address=0x00, data=b"0x01"),
            SupportedFeature.ONBOARD_PROFILES,
        ),
        (
            HIDPPNotification(0, 0, sub_id=Notification.CONNECT_DISCONNECT, address=0x20, data=b"0x01"),
            SupportedFeature.ONBOARD_PROFILES,
        ),
        (
            HIDPPNotification(0, 0, sub_id=Notification.CONNECT_DISCONNECT, address=0x00, data=b"0x01"),
            SupportedFeature.BRIGHTNESS_CONTROL,
        ),
        (
            HIDPPNotification(0, 0, sub_id=Notification.CONNECT_DISCONNECT, address=0x10, data=b"0x01"),
            SupportedFeature.BRIGHTNESS_CONTROL,
        ),
        (
            HIDPPNotification(0, 0, sub_id=Notification.CONNECT_DISCONNECT, address=0x20, data=b"0x01"),
            SupportedFeature.BRIGHTNESS_CONTROL,
        ),
    ],
)
def test_process_feature_notification(mocker, hidpp_notification, feature):
    fake_device = fake_hidpp.Device()
    fake_device.receiver = ["rec1", "rec2"]

    result = notifications._process_feature_notification(fake_device, hidpp_notification)

    assert result is True


def test_process_receiver_notification_invalid(mocker):
    invalid_sub_id = 0x30
    notification_data = b"\x02"
    notification = HIDPPNotification(0, 0, invalid_sub_id, 0, notification_data)
    mock_receiver = mocker.Mock()

    with pytest.raises(AssertionError):
        notifications.process_receiver_notification(mock_receiver, notification)


@pytest.mark.parametrize(
    "sub_id, notification_data, expected",
    [
        (Notification.NO_OPERATION, b"\x00", False),
    ],
)
def test_process_device_notification_extended(mocker, sub_id, notification_data, expected):
    device = mocker.Mock()
    device.handle_notification.return_value = None
    device.protocol = 2.0
    notification = HIDPPNotification(0, 0, sub_id, 0, notification_data)

    result = notifications.process_device_notification(device, notification)

    assert result == expected


def test_handle_device_discovery():
    receiver: Receiver = Receiver(MockLowLevelInterface(), None, {}, True, None, None)
    sub_id = Registers.DISCOVERY_STATUS_NOTIFICATION
    data = b"\x01\x02\x03\x04\x05\x06"
    notification = HIDPPNotification(0, 0, sub_id, 0, data)

    result = notifications.handle_device_discovery(receiver, notification)

    assert result


def test_handle_passkey_request(mocker):
    receiver_mock = mocker.Mock()
    data = b"\x01"
    notification = HIDPPNotification(0, 0, 0, 0, data)

    result = notifications.handle_passkey_request(receiver_mock, notification)

    assert result is True


def test_handle_passkey_pressed(mocker):
    receiver = mocker.Mock()
    sub_id = Registers.DISCOVERY_STATUS_NOTIFICATION
    data = b"\x01\x02\x03\x04\x05\x06"
    notification = HIDPPNotification(0, 0, sub_id, 0, data)

    result = notifications.handle_passkey_pressed(receiver, notification)

    assert result is True



## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""Generic Human Interface Device API.

It is currently a partial pure-Python implementation of the native HID API
implemented by signal11 (https://github.com/signal11/hidapi), and requires
``pyudev``.
The docstrings are mostly copied from the hidapi API header, with changes where
necessary.
"""

from __future__ import annotations

import errno
import logging
import os
import typing
import warnings


# the tuple object we'll expose when enumerating devices
from select import select
from time import sleep
from time import time
from typing import Callable

import pyudev

from hidapi.common import DeviceInfo

if typing.TYPE_CHECKING:
    import gi

    gi.require_version("Gdk", "3.0")
    from gi.repository import GLib  # NOQA: E402

logger = logging.getLogger(__name__)

fileopen = open

ACTION_ADD = "add"
ACTION_REMOVE = "remove"

#
# exposed API
# docstrings mostly copied from hidapi.h
#


def init():
    """This function is a no-op, and exists only to match the native hidapi
    implementation.

    :returns: ``True``.
    """
    return True


def exit():
    """This function is a no-op, and exists only to match the native hidapi
    implementation.

    :returns: ``True``.
    """
    return True


def _match(action: str, device, filter_func: typing.Callable[[int, int, int, bool, bool], dict[str, typing.Any]]):
    """

    The filter_func is used to determine whether this is a device of
    interest to Solaar. It is given the bus id, vendor id, and product
    id and returns a dictionary with the required hid_driver and
    usb_interface and whether this is a receiver or device."""
    logger.debug(f"Dbus event {action} {device}")
    hid_device = device.find_parent("hid")
    if hid_device is None:  # only HID devices are of interest to Solaar
        return
    hid_id = hid_device.properties.get("HID_ID")
    if not hid_id:
        return  # there are reports that sometimes the id isn't set up right so be defensive
    bid, vid, pid = hid_id.split(":")
    hid_hid_device = hid_device.find_parent("hid")
    if hid_hid_device is not None:
        return  # these are devices connected through a receiver so don't pick them up here

    try:  # if report descriptor does not indicate HID++ capabilities then this device is not of interest to Solaar
        from hid_parser import ReportDescriptor

        hidpp_short = hidpp_long = False
        devfile = "/sys" + hid_device.properties.get("DEVPATH") + "/report_descriptor"
        with fileopen(devfile, "rb") as fd:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                rd = ReportDescriptor(fd.read())
            hidpp_short = 0x10 in rd.input_report_ids and 6 * 8 == int(rd.get_input_report_size(0x10))
            # and _Usage(0xFF00, 0x0001) in rd.get_input_items(0x10)[0].usages  # be more permissive
            hidpp_long = 0x11 in rd.input_report_ids and 19 * 8 == int(rd.get_input_report_size(0x11))
            # and _Usage(0xFF00, 0x0002) in rd.get_input_items(0x11)[0].usages  # be more permissive
        if not hidpp_short and not hidpp_long:
            return
    except Exception as e:  # if can't process report descriptor fall back to old scheme
        hidpp_short = None
        hidpp_long = None
        logger.info(
            "Report Descriptor not processed for DEVICE %s BID %s VID %s PID %s: %s",
            device.device_node,
            bid,
            vid,
            pid,
            e,
        )

    filtered_result = filter_func(int(bid, 16), int(vid, 16), int(pid, 16), hidpp_short, hidpp_long)
    if not filtered_result:
        return
    interface_number = filtered_result.get("usb_interface")
    isDevice = filtered_result.get("isDevice")

    if action == ACTION_ADD:
        hid_driver_name = hid_device.properties.get("DRIVER")
        intf_device = device.find_parent("usb", "usb_interface")
        usb_interface = None if intf_device is None else intf_device.attributes.asint("bInterfaceNumber")
        # print('*** usb interface', action, device, 'usb_interface:', intf_device, usb_interface, interface_number)
        logger.info(
            "Found device %s BID %s VID %s PID %s HID++ %s %s USB %s %s",
            device.device_node,
            bid,
            vid,
            pid,
            hidpp_short,
            hidpp_long,
            usb_interface,
            interface_number,
        )
        if not (hidpp_short or hidpp_long or interface_number is None or interface_number == usb_interface):
            return
        attrs = intf_device.attributes if intf_device is not None else None

        d_info = DeviceInfo(
            path=device.device_node,
            bus_id=int(bid, 16),
            vendor_id=vid[-4:],
            product_id=pid[-4:],
            interface=usb_interface,
            driver=hid_driver_name,
            manufacturer=attrs.get("manufacturer") if attrs else None,
            product=attrs.get("product") if attrs else None,
            serial=hid_device.properties.get("HID_UNIQ"),
            release=attrs.get("bcdDevice") if attrs else None,
            isDevice=isDevice,
            hidpp_short=hidpp_short,
            hidpp_long=hidpp_long,
        )
        return d_info

    elif action == ACTION_REMOVE:
        d_info = DeviceInfo(
            path=device.device_node,
            bus_id=None,
            vendor_id=vid[-4:],
            product_id=pid[-4:],
            interface=None,
            driver=None,
            manufacturer=None,
            product=None,
            serial=None,
            release=None,
            isDevice=isDevice,
            hidpp_short=None,
            hidpp_long=None,
        )
        return d_info


def find_paired_node(receiver_path: str, index: int, timeout: int):
    """Find the node of a device paired with a receiver"""
    context = pyudev.Context()
    receiver_phys = pyudev.Devices.from_device_file(context, receiver_path).find_parent("hid").get("HID_PHYS")

    if not receiver_phys:
        return None

    phys = f"{receiver_phys}:{index}"  # noqa: E231
    timeout += time()
    delta = time()
    while delta < timeout:
        for dev in context.list_devices(subsystem="hidraw"):
            dev_phys = dev.find_parent("hid").get("HID_PHYS")
            if dev_phys and dev_phys == phys:
                return dev.device_node
        delta = time()

    return None


def find_paired_node_wpid(receiver_path: str, index: int):
    """Find the node of a device paired with a receiver, get wpid from udev"""
    context = pyudev.Context()
    receiver_phys = pyudev.Devices.from_device_file(context, receiver_path).find_parent("hid").get("HID_PHYS")

    if not receiver_phys:
        return None

    phys = f"{receiver_phys}:{index}"  # noqa: E231
    for dev in context.list_devices(subsystem="hidraw"):
        dev_phys = dev.find_parent("hid").get("HID_PHYS")
        if dev_phys and dev_phys == phys:
            # get hid id like 0003:0000046D:00000065
            hid_id = dev.find_parent("hid").get("HID_ID")
            # get wpid - last 4 symbols
            udev_wpid = hid_id[-4:]
            return udev_wpid

    return None


def monitor_glib(glib: GLib, callback: Callable, filter_func: Callable):
    """Monitor GLib.

    Parameters
    ----------
    glib
        GLib instance.
    """
    c = pyudev.Context()
    m = pyudev.Monitor.from_netlink(c)
    m.filter_by(subsystem="hidraw")

    def _process_udev_event(monitor, condition, cb, filter_func):
        if condition == glib.IO_IN:
            event = monitor.receive_device()
            if event:
                action, device = event
                # print ("***", action, device)
                if action == ACTION_ADD:
                    d_info = _match(action, device, filter_func)
                    if d_info:
                        glib.idle_add(cb, action, d_info)
                elif action == ACTION_REMOVE:
                    # the GLib notification does _not_ match!
                    pass
        return True

    try:
        # io_add_watch_full may not be available...
        glib.io_add_watch_full(m, glib.PRIORITY_LOW, glib.IO_IN, _process_udev_event, callback, filter_func)
    except AttributeError:
        try:
            # and the priority parameter appeared later in the API
            glib.io_add_watch(m, glib.PRIORITY_LOW, glib.IO_IN, _process_udev_event, callback, filter_func)
        except Exception:
            glib.io_add_watch(m, glib.IO_IN, _process_udev_event, callback, filter_func)

    logger.debug("Starting dbus monitoring")
    m.start()


def enumerate(filter_func: typing.Callable[[int, int, int, bool, bool], dict[str, typing.Any]]):
    """Enumerate the HID Devices.

    List all the HID devices attached to the system, optionally filtering by
    vendor_id, product_id, and/or interface_number.

    :returns: a list of matching ``DeviceInfo`` tuples.
    """

    logger.debug("Starting dbus enumeration")
    for dev in pyudev.Context().list_devices(subsystem="hidraw"):
        dev_info = _match(ACTION_ADD, dev, filter_func)
        if dev_info:
            yield dev_info


def open(vendor_id, product_id, serial=None):
    """Open a HID device by its Vendor ID, Product ID and optional serial number.

    If no serial is provided, the first device with the specified IDs is opened.

    :returns: an opaque device handle, or ``None``.
    """

    def matchfn(bid, vid, pid):
        return vid == vendor_id and pid == product_id

    for device in enumerate(matchfn):
        if serial is None or serial == device.serial:
            return open_path(device.path)


def open_path(device_path):
    """Open a HID device by its path name.

    :param device_path: the path of a ``DeviceInfo`` tuple returned by enumerate().

    :returns: an opaque device handle, or ``None``.
    """
    assert device_path
    assert device_path.startswith("/dev/hidraw")

    logger.info("OPEN PATH %s", device_path)
    retrycount = 0
    while retrycount < 3:
        retrycount += 1
        try:
            return os.open(device_path, os.O_RDWR | os.O_SYNC)
        except OSError as e:
            logger.info("OPEN PATH FAILED %s ERROR %s %s", device_path, e.errno, e)
            if e.errno == errno.EACCES:
                sleep(0.1)
            else:
                raise e


def close(device_handle) -> None:
    """Close a HID device.

    :param device_handle: a device handle returned by open() or open_path().
    """
    assert device_handle
    os.close(device_handle)


def write(device_handle, data):
    """Write an Output report to a HID device.

    :param device_handle: a device handle returned by open() or open_path().
    :param data: the data bytes to send including the report number as the
    first byte.

    The first byte of data[] must contain the Report ID. For
    devices which only support a single report, this must be set
    to 0x0. The remaining bytes contain the report data. Since
    the Report ID is mandatory, calls to hid_write() will always
    contain one more byte than the report contains. For example,
    if a hid report is 16 bytes long, 17 bytes must be passed to
    hid_write(), the Report ID (or 0x0, for devices with a
    single report), followed by the report data (16 bytes). In
    this example, the length passed in would be 17.

    write() will send the data on the first OUT endpoint, if
    one exists. If it does not, it will send the data through
    the Control Endpoint (Endpoint 0).
    """
    assert device_handle
    assert data
    assert isinstance(data, bytes), (repr(data), type(data))
    retrycount = 0
    bytes_written = 0
    while retrycount < 3:
        try:
            retrycount += 1
            bytes_written = os.write(device_handle, data)
        except OSError as e:
            if e.errno == errno.EPIPE:
                sleep(0.1)
        else:
            break
    if bytes_written != len(data):
        raise OSError(errno.EIO, f"written {int(bytes_written)} bytes out of expected {len(data)}")


def read(device_handle, bytes_count, timeout_ms=-1):
    """Read an Input report from a HID device.

    :param device_handle: a device handle returned by open() or open_path().
    :param bytes_count: maximum number of bytes to read.
    :param timeout_ms: can be -1 (default) to wait for data indefinitely, 0 to
    read whatever is in the device's input buffer, or a positive integer to
    wait that many milliseconds.

    Input reports are returned to the host through the INTERRUPT IN endpoint.
    The first byte will contain the Report number if the device uses numbered
    reports.

    :returns: the data packet read, an empty bytes string if a timeout was
    reached, or None if there was an error while reading.
    """
    assert device_handle
    timeout = None if timeout_ms < 0 else timeout_ms / 1000.0
    rlist, wlist, xlist = select([device_handle], [], [device_handle], timeout)

    if xlist:
        assert xlist == [device_handle]
        raise OSError(errno.EIO, f"exception on file descriptor {int(device_handle)}")

    if rlist:
        assert rlist == [device_handle]
        data = os.read(device_handle, bytes_count)
        assert data is not None
        assert isinstance(data, bytes), (repr(data), type(data))
        return data
    else:
        return b""


_DEVICE_STRINGS = {
    0: "manufacturer",
    1: "product",
    2: "serial",
}


def get_manufacturer(device_handle):
    """Get the Manufacturer String from a HID device.

    :param device_handle: a device handle returned by open() or open_path().
    """
    return get_indexed_string(device_handle, 0)


def get_product(device_handle):
    """Get the Product String from a HID device.

    :param device_handle: a device handle returned by open() or open_path().
    """
    return get_indexed_string(device_handle, 1)


def get_serial(device_handle):
    """Get the serial number from a HID device.

    :param device_handle: a device handle returned by open() or open_path().
    """
    serial = get_indexed_string(device_handle, 2)
    return serial


def get_indexed_string(device_handle, index):
    """Get a string from a HID device, based on its string index.

    Note: currently not working in the ``hidraw`` native implementation.

    :param device_handle: a device handle returned by open() or open_path().
    :param index: the index of the string to get.
    :returns: the value corresponding to index, or None if no value found
    :rtype: bytes or NoneType
    """
    try:
        key = _DEVICE_STRINGS[index]
    except KeyError:
        return None

    assert device_handle
    stat = os.fstat(device_handle)
    try:
        dev = pyudev.Devices.from_device_number(pyudev.Context(), "char", stat.st_rdev)
    except (pyudev.DeviceNotFoundError, ValueError):
        return None

    hid_dev = dev.find_parent("hid")
    if hid_dev:
        assert "HID_ID" in hid_dev
        bus, _ignore, _ignore = hid_dev["HID_ID"].split(":")

        if bus == "0003":  # USB
            usb_dev = dev.find_parent("usb", "usb_device")
            assert usb_dev
            return usb_dev.attributes.get(key)

        elif bus == "0005":  # BLUETOOTH
            # TODO
            pass



import pytest

from logitech_receiver import base_usb
from logitech_receiver.common import LOGITECH_VENDOR_ID


def test_ensure_known_receivers_mappings_are_valid():
    for key, receiver in base_usb.KNOWN_RECEIVERS.items():
        assert key == receiver["product_id"]


def test_get_receiver_info():
    expected = {
        "vendor_id": LOGITECH_VENDOR_ID,
        "product_id": 0xC548,
        "usb_interface": 2,
        "name": "Bolt Receiver",
        "receiver_kind": "bolt",
        "max_devices": 6,
        "may_unpair": True,
    }

    res = base_usb.get_receiver_info(0xC548)

    assert res == expected


def test_get_receiver_info_unknown_device_fails():
    with pytest.raises(ValueError):
        base_usb.get_receiver_info(0xC500)



import pytest

from logitech_receiver import settings_validator


@pytest.mark.parametrize(
    "current, new, expected",
    [
        (False, "toggle", True),
        (True, "~", False),
        ("don't care", True, True),
        ("don't care", "true", True),
        ("don't care", "false", False),
        ("don't care", "no", False),
        ("don't care", "off", False),
        ("don't care", "True", True),
        ("don't care", "yes", True),
        ("don't care", "on", True),
        ("anything", "anything", None),
    ],
)
def test_bool_or_toggle(current, new, expected):
    result = settings_validator.bool_or_toggle(current=current, new=new)

    assert result == expected



from solaar.ui.about import about
from solaar.ui.about.model import AboutModel


def test_about_model():
    expected_name = "Daniel Pavel"
    model = AboutModel()

    authors = model.get_authors()

    assert expected_name in authors[0]


def test_about_dialog(mocker):
    view_mock = mocker.Mock()

    about.show(view=view_mock)

    assert view_mock.init_ui.call_count == 1
    assert view_mock.update_version_info.call_count == 1
    assert view_mock.update_description.call_count == 1
    assert view_mock.update_authors.call_count == 1
    assert view_mock.update_credits.call_count == 1
    assert view_mock.update_copyright.call_count == 1
    assert view_mock.update_translators.call_count == 1
    assert view_mock.update_website.call_count == 1
    assert view_mock.show.call_count == 1



from dataclasses import dataclass
from dataclasses import field
from typing import Any
from typing import List
from typing import Optional

import gi
import pytest

from logitech_receiver import receiver
from solaar.ui import pair_window

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk  # NOQA: E402

gtk_init = Gtk.init_check()[0]


@dataclass
class Device:
    name: str = "test device"
    kind: str = "test kind"


@dataclass
class Receiver:
    name: str
    receiver_kind: str
    _set_lock: bool = True
    pairing: receiver.Pairing = field(default_factory=receiver.Pairing)
    pairable: bool = True
    _remaining_pairings: Optional[int] = None

    def reset_pairing(self):
        self.receiver = receiver.Pairing()

    def remaining_pairings(self, cache=True):
        return self._remaining_pairings

    def set_lock(self, value=False, timeout=0):
        self.pairing.lock_open = self._set_lock
        return self._set_lock

    def discover(self, cancel=False, timeout=30):
        self.pairing.discovering = self._set_lock
        return self._set_lock

    def pair_device(self, pair=True, slot=0, address=b"\0\0\0\0\0\0", authentication=0x00, entropy=20, force=False):
        print("PD", self.pairable)
        return self.pairable


@dataclass
class Assistant:
    drawable: bool = True
    pages: List[Any] = field(default_factory=list)

    def is_drawable(self):
        return self.drawable

    def next_page(self):
        return True

    def set_page_complete(self, page, b):
        return True

    def commit(self):
        return True

    def append_page(self, page):
        self.pages.append(page)

    def remove_page(self, page):
        return True

    def set_page_type(self, page, type):
        return True

    def destroy(self):
        pass


@pytest.mark.skipif(not gtk_init, reason="requires Gtk")
@pytest.mark.parametrize(
    "receiver, lock_open, discovering, page_type",
    [
        (Receiver("unifying", "unifying", True), True, False, Gtk.AssistantPageType.PROGRESS),
        (Receiver("unifying", "unifying", False), False, False, Gtk.AssistantPageType.SUMMARY),
        (Receiver("nano", "nano", True, _remaining_pairings=5), True, False, Gtk.AssistantPageType.PROGRESS),
        (Receiver("nano", "nano", False), False, False, Gtk.AssistantPageType.SUMMARY),
        (Receiver("bolt", "bolt", True), False, True, Gtk.AssistantPageType.PROGRESS),
        (Receiver("bolt", "bolt", False), False, False, Gtk.AssistantPageType.SUMMARY),
    ],
)
def test_create(receiver, lock_open, discovering, page_type):
    assistant = pair_window.create(receiver)

    assert assistant is not None
    assert assistant.get_page_type(assistant.get_nth_page(0)) == page_type

    assert receiver.pairing.lock_open == lock_open
    assert receiver.pairing.discovering == discovering


@pytest.mark.parametrize(
    "receiver, expected_result, expected_error",
    [
        (Receiver("unifying", "unifying", True), True, False),
        (Receiver("unifying", "unifying", False), False, True),
        (Receiver("bolt", "bolt", True), True, False),
        (Receiver("bolt", "bolt", False), False, True),
    ],
)
def test_prepare(receiver, expected_result, expected_error):
    result = pair_window.prepare(receiver)

    assert result == expected_result
    assert bool(receiver.pairing.error) == expected_error


@pytest.mark.parametrize("assistant, expected_result", [(Assistant(True), True), (Assistant(False), False)])
def test_check_lock_state_drawable(assistant, expected_result):
    r = Receiver("succeed", "unifying", True, receiver.Pairing(lock_open=True))

    result = pair_window.check_lock_state(assistant, r, 2)

    assert result == expected_result


@pytest.mark.skipif(not gtk_init, reason="requires Gtk")
@pytest.mark.parametrize(
    "receiver, count, expected_result",
    [
        (Receiver("fail", "unifying", False, receiver.Pairing(lock_open=False)), 2, False),
        (Receiver("succeed", "unifying", True, receiver.Pairing(lock_open=True)), 1, True),
        (Receiver("error", "unifying", True, receiver.Pairing(error="error")), 0, False),
        (Receiver("new device", "unifying", True, receiver.Pairing(new_device=Device())), 2, False),
        (Receiver("closed", "unifying", True, receiver.Pairing()), 2, False),
        (Receiver("closed", "unifying", True, receiver.Pairing()), 1, False),
        (Receiver("closed", "unifying", True, receiver.Pairing()), 0, False),
        (Receiver("fail bolt", "bolt", False), 1, False),
        (Receiver("succeed bolt", "bolt", True, receiver.Pairing(lock_open=True)), 0, True),
        (Receiver("error bolt", "bolt", True, receiver.Pairing(error="error")), 2, False),
        (Receiver("new device", "bolt", True, receiver.Pairing(lock_open=True, new_device=Device())), 1, False),
        (Receiver("discovering", "bolt", True, receiver.Pairing(lock_open=True)), 1, True),
        (Receiver("closed", "bolt", True, receiver.Pairing()), 2, False),
        (Receiver("closed", "bolt", True, receiver.Pairing()), 1, False),
        (Receiver("closed", "bolt", True, receiver.Pairing()), 0, False),
        (
            Receiver(
                "pass1",
                "bolt",
                True,
                receiver.Pairing(lock_open=True, device_passkey=50, device_authentication=0x01),
            ),
            0,
            True,
        ),
        (
            Receiver(
                "pass2",
                "bolt",
                True,
                receiver.Pairing(lock_open=True, device_passkey=50, device_authentication=0x02),
            ),
            0,
            True,
        ),
        (
            Receiver(
                "adt",
                "bolt",
                True,
                receiver.Pairing(discovering=True, device_address=2, device_name=5),
                pairable=True,
            ),
            2,
            True,
        ),
        (
            Receiver(
                "adf",
                "bolt",
                True,
                receiver.Pairing(discovering=True, device_address=2, device_name=5),
                pairable=False,
            ),
            2,
            False,
        ),
        (Receiver("add fail", "bolt", False, receiver.Pairing(device_address=2, device_passkey=5)), 2, False),
    ],
)
def test_check_lock_state(receiver, count, expected_result):
    assistant = Assistant(True)

    check_state = pair_window._check_lock_state(assistant, receiver, count)

    assert check_state == expected_result


@pytest.mark.parametrize(
    "receiver, pair_device, set_lock, discover, error",
    [
        (
            Receiver("unifying", "unifying", pairing=receiver.Pairing(lock_open=False, error="error")),
            0,
            0,
            0,
            None,
        ),
        (
            Receiver("unifying", "unifying", pairing=receiver.Pairing(lock_open=True, error="error")),
            0,
            1,
            0,
            "error",
        ),
        (Receiver("bolt", "bolt", pairing=receiver.Pairing(lock_open=False, error="error")), 0, 0, 0, None),
        (Receiver("bolt", "bolt", pairing=receiver.Pairing(lock_open=True, error="error")), 1, 0, 0, "error"),
        (Receiver("bolt", "bolt", pairing=receiver.Pairing(discovering=True, error="error")), 0, 0, 1, "error"),
    ],
)
def test_finish(receiver, pair_device, set_lock, discover, error, mocker):
    spy_pair_device = mocker.spy(receiver, "pair_device")
    spy_set_lock = mocker.spy(receiver, "set_lock")
    spy_discover = mocker.spy(receiver, "discover")
    assistant = Assistant(True)

    pair_window._finish(assistant, receiver)

    assert spy_pair_device.call_count == pair_device
    assert spy_set_lock.call_count == set_lock
    assert spy_discover.call_count == discover
    assert receiver.pairing.error == error


@pytest.mark.skipif(not gtk_init, reason="requires Gtk")
@pytest.mark.parametrize("error", ["timeout", "device not supported", "too many devices"])
def test_create_failure_page(error, mocker):
    spy_create = mocker.spy(pair_window, "_create_page")

    pair_window._pairing_failed(Assistant(True), Receiver("nano", "nano"), error)

    assert spy_create.call_count == 1



## Copyright (C) 2024  Solaar Contributors https://pwr-solaar.github.io/Solaar/
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import pytest
import yaml

from logitech_receiver import common
from logitech_receiver import exceptions
from logitech_receiver import hidpp20
from logitech_receiver import hidpp20_constants
from logitech_receiver import special_keys
from logitech_receiver.hidpp20 import KeyFlag
from logitech_receiver.hidpp20 import MappingFlag
from logitech_receiver.hidpp20_constants import GestureId

from . import fake_hidpp

_hidpp20 = hidpp20.Hidpp20()

device_offline = fake_hidpp.Device("REGISTERS", False)
device_registers = fake_hidpp.Device("OFFLINE", True, 1.0)
device_nofeatures = fake_hidpp.Device("NOFEATURES", True, 4.5)
device_zerofeatures = fake_hidpp.Device("ZEROFEATURES", True, 4.5, [fake_hidpp.Response("0000", 0x0000, "0001")])
device_broken = fake_hidpp.Device(
    "BROKEN", True, 4.5, [fake_hidpp.Response("0500", 0x0000, "0001"), fake_hidpp.Response(None, 0x0100)]
)
device_standard = fake_hidpp.Device("STANDARD", True, 4.5, fake_hidpp.r_keyboard_2)


@pytest.mark.parametrize(
    "device, expected_result, expected_count",
    [
        (device_offline, False, 0),
        (device_registers, False, 0),
        (device_nofeatures, False, 0),
        (device_zerofeatures, False, 0),
        (device_broken, False, 0),
        (device_standard, True, 9),
    ],
)
def test_FeaturesArray_check(device, expected_result, expected_count):
    featuresarray = hidpp20.FeaturesArray(device)

    result = featuresarray._check()
    result2 = featuresarray._check()

    assert result == expected_result
    assert result2 == expected_result
    assert (hidpp20_constants.SupportedFeature.ROOT in featuresarray) == expected_result
    assert len(featuresarray) == expected_count
    assert bool(featuresarray) == expected_result


@pytest.mark.parametrize(
    "device, expected0, expected1, expected2, expected5, expected5v",
    [
        (device_zerofeatures, None, None, None, None, None),
        (device_standard, 0x0000, 0x0001, 0x0020, hidpp20_constants.SupportedFeature.REPROG_CONTROLS_V4, 3),
    ],
)
def test_FeaturesArray_get_feature(device, expected0, expected1, expected2, expected5, expected5v):
    featuresarray = hidpp20.FeaturesArray(device)
    device.features = featuresarray

    result0 = featuresarray.get_feature(0)
    result1 = featuresarray.get_feature(1)
    result2 = featuresarray.get_feature(2)
    result5 = featuresarray.get_feature(5)
    result2r = featuresarray.get_feature(2)
    result5v = featuresarray.get_feature_version(hidpp20_constants.SupportedFeature.REPROG_CONTROLS_V4)

    assert result0 == expected0
    assert result1 == expected1
    assert result2 == expected2
    assert result2r == expected2
    assert result5 == expected5
    assert result5v == expected5v


@pytest.mark.parametrize(
    "device, expected_result",
    [
        (device_zerofeatures, []),
        (
            device_standard,
            [
                (hidpp20_constants.SupportedFeature.ROOT, 0),
                (hidpp20_constants.SupportedFeature.FEATURE_SET, 1),
                (hidpp20_constants.SupportedFeature.CONFIG_CHANGE, 2),
                (hidpp20_constants.SupportedFeature.DEVICE_FW_VERSION, 3),
                (common.NamedInt(256, "unknown:0100"), 4),
                (hidpp20_constants.SupportedFeature.REPROG_CONTROLS_V4, 5),
                (None, 6),
                (None, 7),
                (hidpp20_constants.SupportedFeature.BATTERY_STATUS, 8),
            ],
        ),
    ],
)
def test_FeaturesArray_enumerate(device, expected_result):
    featuresarray = hidpp20.FeaturesArray(device)

    result = list(featuresarray.enumerate())

    assert result == expected_result


def test_FeaturesArray_setitem():
    featuresarray = hidpp20.FeaturesArray(device_standard)

    featuresarray[hidpp20_constants.SupportedFeature.ROOT] = 3
    featuresarray[hidpp20_constants.SupportedFeature.FEATURE_SET] = 5
    featuresarray[hidpp20_constants.SupportedFeature.FEATURE_SET] = 4

    assert featuresarray[hidpp20_constants.SupportedFeature.FEATURE_SET] == 4
    assert featuresarray.inverse[4] == hidpp20_constants.SupportedFeature.FEATURE_SET


def test_FeaturesArray_delitem():
    featuresarray = hidpp20.FeaturesArray(device_standard)

    with pytest.raises(ValueError):
        del featuresarray[5]


@pytest.mark.parametrize(
    "device, expected0, expected1, expected2, expected1v",
    [(device_zerofeatures, None, None, None, None), (device_standard, 0, 5, None, 3)],
)
def test_FeaturesArray_getitem(device, expected0, expected1, expected2, expected1v):
    featuresarray = hidpp20.FeaturesArray(device)
    device.features = featuresarray

    result_get0 = featuresarray[hidpp20_constants.SupportedFeature.ROOT]
    result_get1 = featuresarray[hidpp20_constants.SupportedFeature.REPROG_CONTROLS_V4]
    result_get2 = featuresarray[hidpp20_constants.SupportedFeature.GKEY]
    result_1v = featuresarray.get_feature_version(hidpp20_constants.SupportedFeature.REPROG_CONTROLS_V4)

    assert result_get0 == expected0
    assert result_get1 == expected1
    assert result_get2 == expected2
    assert result_1v == expected1v


@pytest.mark.parametrize(
    "device, index, cid, task_id, flags, default_task, expected_flags",
    [
        (device_standard, 2, 1, 1, 0x30, "Volume Up", KeyFlag.REPROGRAMMABLE | KeyFlag.DIVERTABLE),
        (device_standard, 1, 2, 2, 0x20, "Volume Down", KeyFlag.DIVERTABLE),
    ],
)
def test_reprogrammable_key_key(device, index, cid, task_id, flags, default_task, expected_flags):
    key = hidpp20.ReprogrammableKey(device, index, cid, task_id, flags)

    assert key._device == device
    assert key.index == index
    assert key._cid == cid
    assert key._tid == task_id
    assert key._flags == flags
    assert key.key == special_keys.CONTROL[cid]
    assert key.default_task == common.NamedInt(cid, default_task)
    assert key.flags == expected_flags


@pytest.mark.parametrize(
    "device, index, cid, task_id, flags, pos, group, gmask, default_task, expected_flags, group_names",
    [
        (
            device_standard,
            1,
            0x51,
            0x39,
            0x60,
            0,
            1,
            1,
            "Right Click",
            KeyFlag.DIVERTABLE | KeyFlag.PERSISTENTLY_DIVERTABLE,
            ["g1"],
        ),
        (
            device_standard,
            2,
            0x52,
            0x3A,
            0x11,
            1,
            2,
            3,
            "Mouse Middle Button",
            KeyFlag.MSE | KeyFlag.REPROGRAMMABLE,
            ["g1", "g2"],
        ),
        (
            device_standard,
            3,
            0x53,
            0x3C,
            0x110,
            2,
            2,
            7,
            "Mouse Back Button",
            KeyFlag.REPROGRAMMABLE | KeyFlag.RAW_XY,
            ["g1", "g2", "g3"],
        ),
    ],
)
def test_reprogrammable_key_v4_key(
    device, index, cid, task_id, flags, pos, group, gmask, default_task, expected_flags, group_names
):
    key = hidpp20.ReprogrammableKeyV4(device, index, cid, task_id, flags, pos, group, gmask)

    assert key._device == device
    assert key.index == index
    assert key._cid == cid
    assert key._tid == task_id
    assert key._flags == flags
    assert key.pos == pos
    assert key.group == group
    assert key._gmask == gmask
    assert key.key == special_keys.CONTROL[cid]
    assert key.default_task == common.NamedInt(cid, default_task)
    assert key.flags == expected_flags
    assert list(key.group_mask) == group_names


@pytest.mark.parametrize(
    "responses, index, mapped_to, remappable_to, expected_mapping_flags",
    [
        (fake_hidpp.responses_key, 1, "Right Click", common.UnsortedNamedInts(Right_Click=81, Left_Click=80), MappingFlag(0)),
        (fake_hidpp.responses_key, 2, "Left Click", None, MappingFlag.DIVERTED),
        (fake_hidpp.responses_key, 3, "Mouse Back Button", None, MappingFlag.DIVERTED | MappingFlag.PERSISTENTLY_DIVERTED),
        (fake_hidpp.responses_key, 4, "Mouse Forward Button", None, MappingFlag.DIVERTED | MappingFlag.RAW_XY_DIVERTED),
    ],
)
# these fields need access all the key data, so start by setting up a device and its key data
def test_reprogrammable_key_v4_query(responses, index, mapped_to, remappable_to, expected_mapping_flags):
    device = fake_hidpp.Device(
        "KEY", responses=responses, feature=hidpp20_constants.SupportedFeature.REPROG_CONTROLS_V4, offset=5
    )
    device._keys = _hidpp20.get_keys(device)

    key = device.keys[index]

    assert key.mapped_to == mapped_to
    assert (key.remappable_to == remappable_to) or remappable_to is None
    assert key.mapping_flags == expected_mapping_flags


@pytest.mark.parametrize(
    "responses, index, diverted, persistently_diverted, rawXY_reporting, remap, sets",
    [
        (fake_hidpp.responses_key, 1, True, False, True, 0x52, ["0051080000"]),
        (fake_hidpp.responses_key, 2, False, True, False, 0x51, ["0052020000", "0052200000", "0052000051"]),
        (fake_hidpp.responses_key, 3, False, True, True, 0x50, ["0053020000", "00530C0000", "0053300000", "0053000050"]),
        (fake_hidpp.responses_key, 4, False, False, False, 0x50, ["0056020000", "0056080000", "0056200000", "0056000050"]),
    ],
)
def test_reprogrammable_key_v4_set(responses, index, diverted, persistently_diverted, rawXY_reporting, remap, sets, mocker):
    responses += [fake_hidpp.Response(r, 0x530, r) for r in sets]
    device = fake_hidpp.Device(
        "KEY", responses=responses, feature=hidpp20_constants.SupportedFeature.REPROG_CONTROLS_V4, offset=5
    )
    device._keys = _hidpp20.get_keys(device)
    device._keys._ensure_all_keys_queried()  # do this now so that the last requests are sets
    spy_request = mocker.spy(device, "request")

    key = device.keys[index]
    _mapping_flags = key.mapping_flags

    if hidpp20.KeyFlag.DIVERTABLE in key.flags or not diverted:
        key.set_diverted(diverted)
    else:
        with pytest.raises(exceptions.FeatureNotSupported):
            key.set_diverted(diverted)
    assert (MappingFlag.DIVERTED in key.mapping_flags) == (diverted and hidpp20.KeyFlag.DIVERTABLE in key.flags)

    if hidpp20.KeyFlag.PERSISTENTLY_DIVERTABLE in key.flags or not persistently_diverted:
        key.set_persistently_diverted(persistently_diverted)
    else:
        with pytest.raises(exceptions.FeatureNotSupported):
            key.set_persistently_diverted(persistently_diverted)
    assert (hidpp20.MappingFlag.PERSISTENTLY_DIVERTED in key.mapping_flags) == (
        persistently_diverted and hidpp20.KeyFlag.PERSISTENTLY_DIVERTABLE in key.flags
    )

    if hidpp20.KeyFlag.RAW_XY in key.flags or not rawXY_reporting:
        key.set_rawXY_reporting(rawXY_reporting)
    else:
        with pytest.raises(exceptions.FeatureNotSupported):
            key.set_rawXY_reporting(rawXY_reporting)
    assert (MappingFlag.RAW_XY_DIVERTED in key.mapping_flags) == (rawXY_reporting and hidpp20.KeyFlag.RAW_XY in key.flags)

    if remap in key.remappable_to or remap == 0:
        key.remap(remap)
    else:
        with pytest.raises(exceptions.FeatureNotSupported):
            key.remap(remap)
    assert (key.mapped_to == remap) or (remap not in key.remappable_to and remap != 0)

    fake_hidpp.match_requests(len(sets), responses, spy_request.call_args_list)


@pytest.mark.parametrize(
    "r, index, cid, actionId, remapped, mask, status, action, modifiers, byts, remap",
    [
        (fake_hidpp.responses_key, 1, 0x0051, 0x02, 0x0002, 0x01, 0, "Mouse Button: 2", "Cntrl+", "02000201", "01000400"),
        (fake_hidpp.responses_key, 2, 0x0052, 0x01, 0x0001, 0x00, 1, "Key: 1", "", "01000100", "02005004"),
        (fake_hidpp.responses_key, 3, 0x0053, 0x02, 0x0001, 0x00, 1, "Mouse Button: 1", "", "02000100", "7FFFFFFF"),
    ],
)
def test_remappable_action(r, index, cid, actionId, remapped, mask, status, action, modifiers, byts, remap, mocker):
    if int(remap, 16) == special_keys.KEYS_Default:
        responses = r + [
            fake_hidpp.Response("040000", 0x0000, "1C00"),
            fake_hidpp.Response("00", 0x450, f"{cid:04X}" + "FF"),
        ]
    else:
        responses = r + [
            fake_hidpp.Response("040000", 0x0000, "1C00"),
            fake_hidpp.Response("00", 0x440, f"{cid:04X}" + "FF" + remap),
        ]
    device = fake_hidpp.Device(
        "KEY", responses=responses, feature=hidpp20_constants.SupportedFeature.REPROG_CONTROLS_V4, offset=5
    )
    key = hidpp20.PersistentRemappableAction(device, index, cid, actionId, remapped, mask, status)
    spy_request = mocker.spy(device, "request")

    assert key._device == device
    assert key.index == index
    assert key._cid == cid
    assert key.actionId == actionId
    assert key.remapped == remapped
    assert key._modifierMask == mask
    assert key.cidStatus == status
    assert key.key == special_keys.CONTROL[cid]
    assert key.actionType == special_keys.ACTIONID[actionId]
    assert key.action == action
    assert key.modifiers == modifiers
    assert key.data_bytes.hex().upper() == byts

    key.remap(bytes.fromhex(remap))
    assert key.data_bytes.hex().upper() == (byts if int(remap, 16) == special_keys.KEYS_Default else remap)

    if int(remap, 16) != special_keys.KEYS_Default:
        fake_hidpp.match_requests(1, responses, spy_request.call_args_list)


# KeysArray methods tested in KeysArrayV4

# KeysArrayV2 not tested as there is no documentation


@pytest.mark.parametrize(
    "device, index", [(device_zerofeatures, -1), (device_zerofeatures, 5), (device_standard, -1), (device_standard, 6)]
)
def test_KeysArrayV4_index_error(device, index):
    keysarray = hidpp20.KeysArrayV4(device, 5)

    with pytest.raises(IndexError):
        keysarray[index]

    with pytest.raises(IndexError):
        keysarray._query_key(index)


@pytest.mark.parametrize("device, index, top, cid", [(device_standard, 0, 2, 0x0011), (device_standard, 4, 5, 0x0003)])
def test_KeysArrayV4_query_key(device, index, top, cid):
    keysarray = hidpp20.KeysArrayV4(device, 5)

    keysarray._query_key(index)

    assert keysarray.keys[index]._cid == cid
    assert len(keysarray[index:top]) == top - index
    assert len(list(keysarray)) == 5


@pytest.mark.parametrize(
    "device, count, index, cid, task_id, flags, pos, group, gmask",
    [
        (device_standard, 4, 0, 0x0011, 0x0012, 0xCDAB, 1, 2, 3),
        (device_standard, 6, 1, 0x0111, 0x0022, 0xCDAB, 1, 2, 3),
        (device_standard, 8, 3, 0x0311, 0x0032, 0xCDAB, 1, 2, 4),
    ],
)
def test_KeysArrayV4__getitem(device, count, index, cid, task_id, flags, pos, group, gmask):
    keysarray = hidpp20.KeysArrayV4(device, count)

    result = keysarray[index]

    assert result._device == device
    assert result.index == index
    assert result._cid == cid
    assert result._tid == task_id
    assert result._flags == flags
    assert result.pos == pos
    assert result.group == group
    assert result._gmask == gmask


@pytest.mark.parametrize(
    "key, index", [(special_keys.CONTROL.Volume_Up_old, 2), (special_keys.CONTROL.Mute, 4), (special_keys.CONTROL.Next, None)]
)
def test_KeysArrayV4_index(key, index):
    keysarray = hidpp20.KeysArrayV4(device_standard, 7)

    result = keysarray.index(key)

    assert result == index


device_key = fake_hidpp.Device(
    "KEY", responses=fake_hidpp.responses_key, feature=hidpp20_constants.SupportedFeature.REPROG_CONTROLS_V4, offset=5
)


@pytest.mark.parametrize(
    "key, expected_index, expected_mapped_to, expected_remappable_to",
    [
        (
            special_keys.CONTROL.Left_Button,
            0,
            common.NamedInt(0x50, "Left Click"),
            [common.NamedInt(0x50, "Left Click"), common.NamedInt(0x51, "Right Click")],
        ),
        (
            special_keys.CONTROL.Right_Button,
            1,
            common.NamedInt(0x51, "Right Click"),
            [common.NamedInt(0x51, "Right Click"), common.NamedInt(0x50, "Left Click")],
        ),
        (special_keys.CONTROL.Middle_Button, 2, common.NamedInt(0x50, "Left Click"), None),
        (special_keys.CONTROL.Back_Button, 3, common.NamedInt(0x53, "Mouse Back Button"), None),
        (special_keys.CONTROL.Forward_Button, 4, common.NamedInt(0x56, "Mouse Forward Button"), None),
        (special_keys.CONTROL.Mouse_Gesture_Button, 5, common.NamedInt(0xC3, "Gesture Button Navigation"), None),
        (special_keys.CONTROL.Smart_Shift, 6, common.NamedInt(0x50, "Left Click"), None),
        (special_keys.CONTROL.Virtual_Gesture_Button, 7, common.NamedInt(0x51, "Right Click"), None),
    ],
)
def test_keys_array_v4_key(key, expected_index, expected_mapped_to, expected_remappable_to):
    device_key._keys = _hidpp20.get_keys(device_key)
    device_key._keys._ensure_all_keys_queried()

    index = device_key._keys.index(key)
    mapped_to = device_key._keys[expected_index].mapped_to
    remappable_to = device_key._keys[expected_index].remappable_to

    assert index == expected_index
    assert mapped_to == expected_mapped_to
    if expected_remappable_to is not None:
        assert list(remappable_to) == expected_remappable_to


@pytest.mark.parametrize(
    "device, index", [(device_zerofeatures, -1), (device_zerofeatures, 5), (device_standard, -1), (device_standard, 6)]
)
def test_KeysArrayPersistent_index_error(device, index):
    keysarray = hidpp20.KeysArrayPersistent(device, 5)

    with pytest.raises(IndexError):
        keysarray[index]

    with pytest.raises(IndexError):
        keysarray._query_key(index)


@pytest.mark.parametrize(
    "responses, key, index, mapped_to, capabilities",
    [
        (fake_hidpp.responses_remap, special_keys.CONTROL.Left_Button, 0, common.NamedInt(0x01, "Mouse Button Left"), 0x41),
        (fake_hidpp.responses_remap, special_keys.CONTROL.Right_Button, 1, common.NamedInt(0x01, "Mouse Button Left"), 0x41),
        (fake_hidpp.responses_remap, special_keys.CONTROL.Middle_Button, 2, common.NamedInt(0x51, "DOWN"), 0x41),
    ],
)
def test_KeysArrayPersistent_key(responses, key, index, mapped_to, capabilities):
    device = fake_hidpp.Device(
        "REMAP", responses=responses, feature=hidpp20_constants.SupportedFeature.PERSISTENT_REMAPPABLE_ACTION
    )
    device._remap_keys = _hidpp20.get_remap_keys(device)
    device._remap_keys._ensure_all_keys_queried()

    assert device._remap_keys.index(key) == index
    assert device._remap_keys[index].remapped == mapped_to
    assert device._remap_keys.capabilities == capabilities


@pytest.mark.parametrize(
    "id, length, minimum, maximum, widget, min, max, wid, string",
    [
        ("left", 1, 5, 8, "Widget", 5, 8, "Widget", "left"),
        ("left", 1, None, None, None, 0, 255, "Scale", "left"),
    ],
)
def test_SubParam(id, length, minimum, maximum, widget, min, max, wid, string):
    subparam = hidpp20.SubParam(id, length, minimum, maximum, widget)

    assert subparam.id == id
    assert subparam.length == length
    assert subparam.minimum == min
    assert subparam.maximum == max
    assert subparam.widget == wid
    assert subparam.__str__() == string
    assert subparam.__repr__() == string


@pytest.mark.parametrize(
    "device, low, high, next_index, next_diversion_index, name, cbe, si, sdi, eom, dom",
    [
        (device_standard, 0x01, 0x01, 5, 10, GestureId.TAP_1_FINGER, True, 5, None, (0, 0x20), (None, None)),
        (device_standard, 0x03, 0x02, 6, 11, GestureId.TAP_3_FINGER, False, None, 11, (None, None), (1, 0x08)),
    ],
)
def test_gesture(device, low, high, next_index, next_diversion_index, name, cbe, si, sdi, eom, dom):
    gesture = hidpp20.Gesture(device, low, high, next_index, next_diversion_index)

    assert gesture._device == device
    assert gesture.id == low
    assert gesture.gesture == name
    assert gesture.can_be_enabled == cbe
    assert gesture.can_be_enabled == cbe
    assert gesture.index == si
    assert gesture.diversion_index == sdi
    assert gesture.enable_offset_mask() == eom
    assert gesture.diversion_offset_mask() == dom
    assert gesture.as_int() == low
    assert int(gesture) == low


@pytest.mark.parametrize(
    "responses, gest, enabled, diverted, set_result, unset_result, divert_result, undivert_result",
    [
        (fake_hidpp.responses_gestures, 20, None, None, None, None, None, None),
        (fake_hidpp.responses_gestures, 1, True, False, "01", "00", "01", "00"),
        (fake_hidpp.responses_gestures, 45, False, None, "01", "00", None, None),
    ],
)
def test_Gesture_set(responses, gest, enabled, diverted, set_result, unset_result, divert_result, undivert_result):
    device = fake_hidpp.Device("GESTURE", responses=responses, feature=hidpp20_constants.SupportedFeature.GESTURE_2)
    gestures = _hidpp20.get_gestures(device)

    gesture = gestures.gesture(gest)

    assert gesture.enabled() == enabled
    assert gesture.diverted() == diverted
    assert gesture.set(True) == (bytes.fromhex(set_result) if set_result is not None else None)
    assert gesture.set(False) == (bytes.fromhex(unset_result) if unset_result is not None else None)
    assert gesture.divert(True) == (bytes.fromhex(divert_result) if divert_result is not None else None)
    assert gesture.divert(False) == (bytes.fromhex(undivert_result) if undivert_result is not None else None)


@pytest.mark.parametrize(
    "responses, prm, id, index, size, value, default_value, write1, write2",
    [
        (fake_hidpp.responses_gestures, 4, hidpp20_constants.ParamId.SCALE_FACTOR, 0, 2, 256, 256, "0080", "0180"),
    ],
)
def test_param(responses, prm, id, index, size, value, default_value, write1, write2):
    device = fake_hidpp.Device("GESTURE", responses=responses, feature=hidpp20_constants.SupportedFeature.GESTURE_2)
    gestures = _hidpp20.get_gestures(device)

    param = gestures.param(prm)

    assert param.id == id
    assert param.index == index
    assert param.size == size
    assert param.value == value
    assert param.default_value == default_value
    assert param.param == id
    assert int(param) == id
    assert param.write(bytes.fromhex(write1)).hex().upper() == f"{index:02X}" + write1 + "FF"
    assert param.write(bytes.fromhex(write2)).hex().upper() == f"{index:02X}" + write2 + "FF"


@pytest.mark.parametrize(
    "responses, id, s, byte_count, expected_value, expected_string",
    [
        (fake_hidpp.responses_gestures, 1, hidpp20.SpecGesture.DVI_FIELD_WIDTH, 1, 8, "[dvi field width=8]"),
        (fake_hidpp.responses_gestures, 2, hidpp20.SpecGesture.FIELD_WIDTHS, 1, 8, "[field widths=8]"),
        (fake_hidpp.responses_gestures, 3, hidpp20.SpecGesture.PERIOD_UNIT, 2, 2048, "[period unit=2048]"),
    ],
)
def test_spec(responses, id, s, byte_count, expected_value, expected_string):
    device = fake_hidpp.Device("GESTURE", responses=responses, feature=hidpp20_constants.SupportedFeature.GESTURE_2)
    gestures = _hidpp20.get_gestures(device)

    spec = gestures.specs[id]

    assert spec.id == id
    assert spec.spec == s
    assert spec.byte_count == byte_count
    assert spec.value == expected_value
    assert repr(spec) == expected_string


def test_Gestures():
    device = fake_hidpp.Device(
        "GESTURES", responses=fake_hidpp.responses_gestures, feature=hidpp20_constants.SupportedFeature.GESTURE_2
    )
    gestures = _hidpp20.get_gestures(device)

    assert gestures

    assert len(gestures.gestures) == 17
    assert gestures.gesture(20) == gestures.gestures[20]
    assert gestures.gesture_enabled(20) is None
    assert gestures.gesture_enabled(1) is True
    assert gestures.gesture_enabled(45) is False
    assert gestures.enable_gesture(20) is None
    assert gestures.enable_gesture(45) == bytes.fromhex("01")
    assert gestures.disable_gesture(20) is None
    assert gestures.disable_gesture(45) == bytes.fromhex("00")

    assert len(gestures.params) == 1
    assert gestures.param(4) == gestures.params[4]
    assert gestures.get_param(4) == 256
    assert gestures.set_param(4, 128) is None

    assert len(gestures.specs) == 5
    assert gestures.specs[2].value == 8
    assert gestures.specs[4].value == 4


responses_backlight = [
    fake_hidpp.Response("010118000001020003000400", 0x0400),
    fake_hidpp.Response("0101FF00020003000400", 0x0410, "0101FF00020003000400"),
]

device_backlight = fake_hidpp.Device(
    "BACKLIGHT", responses=responses_backlight, feature=hidpp20_constants.SupportedFeature.BACKLIGHT2
)


def test_Backlight():
    backlight = _hidpp20.get_backlight(device_backlight)
    result = backlight.write()

    assert backlight
    assert backlight.auto_supported
    assert backlight.temp_supported
    assert not backlight.perm_supported
    assert backlight.dho == 0x0002
    assert backlight.dhi == 0x0003
    assert backlight.dpow == 0x0004
    assert result is not None


@pytest.mark.parametrize(
    "hex, ID, color, speed, period, intensity, ramp, form",
    [
        ("FFFFFFFFFFFFFFFFFFFFFF", None, None, None, None, None, None, None),
        ("0000000000000000000000", common.NamedInt(0x0, "Disabled"), None, None, None, None, None, None),
        ("0120304010000000000000", common.NamedInt(0x1, "Static"), 0x203040, None, None, None, 0x10, None),
        ("0220304010000000000000", common.NamedInt(0x2, "Pulse"), 0x203040, 0x10, None, None, None, None),
        ("0800000000000000000000", common.NamedInt(0x8, "Boot"), None, None, None, None, None, None),
        ("0300000000005000000000", common.NamedInt(0x3, "Cycle"), None, None, 0x5000, 0x00, None, None),
        ("0A20304010005020000000", common.NamedInt(0xA, "Breathe"), 0x203040, None, 0x1000, 0x20, None, 0x50),
        ("0B20304000100000000000", common.NamedInt(0xB, "Ripple"), 0x203040, None, 0x1000, None, None, None),
        ("0A01020300500407000000", common.NamedInt(0xA, "Breathe"), 0x010203, None, 0x0050, 0x07, None, 0x04),
    ],
)
def test_LEDEffectSetting(hex, ID, color, speed, period, intensity, ramp, form):
    byt = bytes.fromhex(hex)
    setting = hidpp20.LEDEffectSetting.from_bytes(byt)

    assert setting.ID == ID
    if ID is None:
        assert setting.bytes == byt
    else:
        assert getattr(setting, "color", None) == color
        assert getattr(setting, "speed", None) == speed
        assert getattr(setting, "period", None) == period
        assert getattr(setting, "intensity", None) == intensity
        assert getattr(setting, "ramp", None) == ramp
        assert getattr(setting, "form", None) == form

    assert setting.to_bytes() == byt
    assert yaml.safe_load(yaml.dump(setting)) == setting
    assert yaml.safe_load(str(setting)) == setting


@pytest.mark.parametrize(
    "feature, function, response, ID, capabilities, period",
    [
        [
            hidpp20_constants.SupportedFeature.COLOR_LED_EFFECTS,
            0x20,
            fake_hidpp.Response("0102000300040005", 0x0420, "010200"),
            3,
            4,
            5,
        ],
        [
            hidpp20_constants.SupportedFeature.COLOR_LED_EFFECTS,
            0x20,
            fake_hidpp.Response("0102000700080009", 0x0420, "010200"),
            7,
            8,
            9,
        ],
    ],
)
def test_LEDEffectInfo(feature, function, response, ID, capabilities, period):
    device = fake_hidpp.Device(feature=feature, responses=[response])

    info = hidpp20.LEDEffectInfo(feature, function, device, 1, 2)

    assert info.zindex == 1
    assert info.index == 2
    assert info.ID == ID
    assert info.capabilities == capabilities
    assert info.period == period


@pytest.mark.parametrize(
    "feature, function, offset, effect_function, responses, index, location, count, id_1",
    [
        [hidpp20_constants.SupportedFeature.COLOR_LED_EFFECTS, 0x10, 0, 0x20, fake_hidpp.zone_responses_1, 0, 1, 2, 0xB],
        [hidpp20_constants.SupportedFeature.RGB_EFFECTS, 0x00, 1, 0x00, fake_hidpp.zone_responses_2, 0, 1, 2, 2],
    ],
)
def test_LEDZoneInfo(feature, function, offset, effect_function, responses, index, location, count, id_1):
    device = fake_hidpp.Device(feature=feature, responses=responses, offset=0x07)

    zone = hidpp20.LEDZoneInfo(feature, function, offset, effect_function, device, index)

    assert zone.index == index
    assert zone.location == location
    assert zone.count == count
    assert len(zone.effects) == count
    assert zone.effects[1].ID == id_1


@pytest.mark.parametrize(
    "responses, setting, expected_command",
    [
        [fake_hidpp.zone_responses_1, hidpp20.LEDEffectSetting(ID=0), None],
        [fake_hidpp.zone_responses_1, hidpp20.LEDEffectSetting(ID=3, period=0x20, intensity=0x50), "000000000000000020500000"],
        [
            fake_hidpp.zone_responses_1,
            hidpp20.LEDEffectSetting(ID=0xB, color=0x808080, period=0x20),
            "000180808000002000000000",
        ],
    ],
)
def test_LEDZoneInfo_to_command(responses, setting, expected_command):
    device = fake_hidpp.Device(feature=hidpp20_constants.SupportedFeature.COLOR_LED_EFFECTS, responses=responses, offset=0x07)
    zone = hidpp20.LEDZoneInfo(hidpp20_constants.SupportedFeature.COLOR_LED_EFFECTS, 0x10, 0, 0x20, device, 0)

    command = zone.to_command(setting)

    assert command == (bytes.fromhex(expected_command) if expected_command is not None else None)


@pytest.mark.parametrize(
    "feature, cls, responses, readable, count, count_0",
    [
        [
            hidpp20_constants.SupportedFeature.COLOR_LED_EFFECTS,
            hidpp20.LEDEffectsInfo,
            fake_hidpp.effects_responses_1,
            1,
            1,
            2,
        ],
        [hidpp20_constants.SupportedFeature.RGB_EFFECTS, hidpp20.RGBEffectsInfo, fake_hidpp.effects_responses_2, 1, 1, 2],
    ],
)
def test_LED_RGB_EffectsInfo(feature, cls, responses, readable, count, count_0):
    device = fake_hidpp.Device(feature=feature, responses=responses, offset=0x07)

    effects = cls(device)

    assert effects.readable == readable
    assert effects.count == count
    assert effects.zones[0].count == count_0


@pytest.mark.parametrize(
    "hex, expected_behavior, sector, address, typ, val, modifiers, data, byt",
    [
        ("05010203", 0x0, 0x501, 0x0203, None, None, None, None, None),
        ("15020304", 0x1, 0x502, 0x0304, None, None, None, None, None),
        ("8000FFFF", 0x8, None, None, 0x00, None, None, None, None),
        ("80010102", 0x8, None, None, 0x01, 0x0102, None, None, None),
        ("80020454", 0x8, None, None, 0x02, 0x54, 0x04, None, None),
        ("80030454", 0x8, None, None, 0x03, 0x0454, None, None, None),
        ("900AFF01", 0x9, None, None, None, 0x0A, None, 0x01, None),
        ("709090A0", 0x7, None, None, None, None, None, None, b"\x70\x90\x90\xa0"),
    ],
)
def test_button_bytes(hex, expected_behavior, sector, address, typ, val, modifiers, data, byt):
    button = hidpp20.Button.from_bytes(bytes.fromhex(hex))

    assert getattr(button, "behavior", None) == expected_behavior
    assert getattr(button, "sector", None) == sector
    assert getattr(button, "address", None) == address
    assert getattr(button, "type", None) == typ
    assert getattr(button, "value", None) == val
    assert getattr(button, "modifiers", None) == modifiers
    assert getattr(button, "data", None) == data
    assert getattr(button, "bytes", None) == byt
    assert button.to_bytes().hex().upper() == hex
    assert yaml.safe_load(yaml.dump(button)).to_bytes().hex().upper() == hex


hex1 = (
    "01010290018003000700140028FFFFFF"
    "FFFF0000000000000000000000000000"
    "8000FFFF900AFF00800204548000FFFF"
    "900AFF00800204548000FFFF900AFF00"
    "800204548000FFFF900AFF0080020454"
    "8000FFFF900AFF00800204548000FFFF"
    "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF"
    "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF"
    "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF"
    "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF"
    "54004500370000000000000000000000"
    "00000000000000000000000000000000"
    "00000000000000000000000000000000"
    "0A01020300500407000000FFFFFFFFFF"
    "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF"
    "FFFFFFFFFFFFFFFFFFFFFFFFFF7C81"
)
hex2 = (
    "01010290018003000700140028FFFFFF"
    "FFFF0000000000000000000000000000"
    "8000FFFF900AFF00800204548000FFFF"
    "900AFF00800204548000FFFF900AFF00"
    "800204548000FFFF900AFF0080020454"
    "8000FFFF900AFF00800204548000FFFF"
    "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF"
    "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF"
    "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF"
    "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF"
    "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF"
    "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF"
    "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF"
    "0A01020300500407000000FFFFFFFFFF"
    "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF"
    "FFFFFFFFFFFFFFFFFFFFFFFFFF27C9"
)
hex3 = (
    "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF"
    "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF"
    "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF"
    "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF"
    "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF"
    "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF"
    "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF"
    "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF"
    "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF"
    "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF"
    "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF"
    "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF"
    "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF"
    "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF"
    "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF"
    "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF"
    "FFFFFFFFFFFFFFFFFFFFFFFFFF2307"
)


@pytest.mark.parametrize(
    "hex, name, sector, enabled, buttons, gbuttons, resolutions, button, lighting",
    [
        (hex1, "TE7", 2, 1, 16, 0, [0x0190, 0x0380, 0x0700, 0x1400, 0x2800], "8000FFFF", "0A01020300500407000000"),
        (hex2, "", 2, 1, 16, 0, [0x0190, 0x0380, 0x0700, 0x1400, 0x2800], "8000FFFF", "0A01020300500407000000"),
        (hex3, "", 2, 1, 16, 0, [0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF], "FFFFFFFF", "FFFFFFFFFFFFFFFFFFFFFF"),
    ],
)
def test_onboard_profile_bytes(hex, name, sector, enabled, buttons, gbuttons, resolutions, button, lighting):
    profile = hidpp20.OnboardProfile.from_bytes(sector, enabled, buttons, gbuttons, bytes.fromhex(hex))

    assert profile.name == name
    assert profile.sector == sector
    assert profile.resolutions == resolutions
    assert profile.buttons[0].to_bytes().hex().upper() == button
    assert profile.lighting[0].to_bytes().hex().upper() == lighting

    assert profile.to_bytes(len(hex) // 2).hex().upper() == hex
    assert yaml.safe_load(yaml.dump(profile)).to_bytes(len(hex) // 2).hex().upper() == hex


@pytest.mark.parametrize(
    "responses, name, count, buttons, gbuttons, sectors, size",
    [
        (fake_hidpp.responses_profiles, "ONB", 1, 2, 2, 1, 254),
        (fake_hidpp.responses_profiles_rom, "ONB", 1, 2, 2, 1, 254),
        (fake_hidpp.responses_profiles_rom_2, "ONB", 1, 2, 2, 1, 254),
    ],
)
def test_onboard_profiles_device(responses, name, count, buttons, gbuttons, sectors, size):
    device = fake_hidpp.Device(
        name, True, 4.5, responses=responses, feature=hidpp20_constants.SupportedFeature.ONBOARD_PROFILES, offset=0x9
    )
    device._profiles = None
    profiles = _hidpp20.get_profiles(device)

    assert profiles
    assert profiles.version == hidpp20.OnboardProfilesVersion
    assert profiles.name == name
    assert profiles.count == count
    assert profiles.buttons == buttons
    assert profiles.gbuttons == gbuttons
    assert profiles.sectors == sectors
    assert profiles.size == size
    assert len(profiles.profiles) == count

    yml_dump = yaml.dump(profiles)
    assert yaml.safe_load(yml_dump).to_bytes().hex() == profiles.to_bytes().hex()



import textwrap

from unittest import mock
from unittest.mock import mock_open

import pytest

from logitech_receiver import diversion
from logitech_receiver.base import HIDPPNotification
from logitech_receiver.hidpp20_constants import SupportedFeature


@pytest.fixture
def rule_config():
    rule_content = """
    %YAML 1.3
    ---
    - MouseGesture: Mouse Left
    - KeyPress:
      - [Control_L, Alt_L, Left]
      - click
    ...
    ---
    - MouseGesture: Mouse Up
    - KeyPress:
      - [Super_L, Up]
      - click
    ...
    ---
    - Test: [thumb_wheel_up, 10]
    - KeyPress:
      - [Control_L, Page_Down]
      - click
    ...
    ---
    """
    return textwrap.dedent(rule_content)


def test_load_rule_config(rule_config):
    expected_rules = [
        [
            diversion.MouseGesture,
            diversion.KeyPress,
        ],
        [diversion.MouseGesture, diversion.KeyPress],
        [diversion.Test, diversion.KeyPress],
    ]

    with mock.patch("builtins.open", new=mock_open(read_data=rule_config)):
        loaded_rules = diversion._load_rule_config(file_path=mock.Mock())

    assert len(loaded_rules.components) == 2  # predefined and user configured rules
    user_configured_rules = loaded_rules.components[0]
    assert isinstance(user_configured_rules, diversion.Rule)

    for components, expected_components in zip(user_configured_rules.components, expected_rules):
        for component, expected_component in zip(components.components, expected_components):
            assert isinstance(component, expected_component)


def test_diversion_rule():
    args = [
        {
            "Rule": [  # Implement problematic keys for Craft and MX Master
                {"Rule": [{"Key": ["Brightness Down", "pressed"]}, {"KeyPress": "XF86_MonBrightnessDown"}]},
                {"Rule": [{"Key": ["Brightness Up", "pressed"]}, {"KeyPress": "XF86_MonBrightnessUp"}]},
            ]
        },
    ]

    rule = diversion.Rule(args)

    assert len(rule.components) == 1
    root_rule = rule.components[0]
    assert isinstance(root_rule, diversion.Rule)

    assert len(root_rule.components) == 2
    for component in root_rule.components:
        assert isinstance(component, diversion.Rule)
        assert len(component.components) == 2

        key = component.components[0]
        assert isinstance(key, diversion.Key)
        key = component.components[1]
        assert isinstance(key, diversion.KeyPress)


def test_key_is_down():
    result = diversion.key_is_down(key=diversion.CONTROL.G2)

    assert result is False


def test_feature():
    expected_data = {"Feature": "CONFIG CHANGE"}

    result = diversion.Feature("CONFIG_CHANGE")

    assert result.data() == expected_data


@pytest.mark.parametrize(
    "feature, data",
    [
        (
            SupportedFeature.REPROG_CONTROLS_V4,
            [0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08],
        ),
        (SupportedFeature.GKEY, [0x01, 0x02, 0x03, 0x04]),
        (SupportedFeature.MKEYS, [0x01, 0x02, 0x03, 0x04]),
        (SupportedFeature.MR, [0x01, 0x02, 0x03, 0x04]),
        (SupportedFeature.THUMB_WHEEL, [0x01, 0x02, 0x03, 0x04, 0x05]),
        (SupportedFeature.DEVICE_UNIT_ID, [0x01, 0x02, 0x03, 0x04, 0x05]),
    ],
)
def test_process_notification(feature, data):
    device_mock = mock.Mock()
    notification = HIDPPNotification(
        report_id=0x01,
        devnumber=1,
        sub_id=0x13,
        address=0x00,
        data=bytes(data),
    )

    diversion.process_notification(device_mock, notification, feature)



from solaar.gtk import create_parser


def test_arg_parse():
    parser = create_parser()
    res = parser.parse_args([])

    assert res.debug == 0
    assert res.hidraw_path is None
    assert res.restart_on_wake_up is False
    assert res.window is None
    assert res.battery_icons is None
    assert res.tray_icon_size is None


def test_arg_parse_debug():
    parser = create_parser()
    res = parser.parse_args(["--debug"])

    assert res.debug == 1


def test_arg_parse_version():
    parser = create_parser()
    res = parser.parse_args(["version"])

    assert res



import platform

from unittest import mock

if platform.system() == "Linux":
    import hidapi.udev_impl as hidapi
else:
    import hidapi.hidapi_impl as hidapi


def test_find_paired_node():
    hidapi.enumerate(mock.Mock())



from unittest import mock

from solaar.ui import desktop_notifications

# depends on external environment, so make some tests dependent on availability


def test_init():
    result = desktop_notifications.init()

    assert result == desktop_notifications.available


def test_uninit():
    assert desktop_notifications.uninit() is None


def test_alert():
    reason = "unknown"
    assert desktop_notifications.alert(reason) is None


class MockDevice(mock.Mock):
    name = "MockDevice"

    def close():
        return True


def test_show():
    dev = MockDevice()
    reason = "unknown"
    available = desktop_notifications.init()

    result = desktop_notifications.show(dev, reason)
    if available:
        assert result is not None
    else:
        assert result is None



## Copyright (C) 2024  Solaar Contributors https://pwr-solaar.github.io/Solaar/
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import pytest

from logitech_receiver import common
from logitech_receiver import hidpp20
from logitech_receiver import hidpp20_constants
from logitech_receiver.hidpp20_constants import SupportedFeature

from . import fake_hidpp

_hidpp20 = hidpp20.Hidpp20()


def test_get_firmware():
    responses = [
        fake_hidpp.Response("02FFFF", 0x0400),
        fake_hidpp.Response("01414243030401000101000102030405", 0x0410, "00"),
        fake_hidpp.Response("02414243030401000101000102030405", 0x0410, "01"),
    ]
    device = fake_hidpp.Device(responses=responses, feature=SupportedFeature.DEVICE_FW_VERSION)

    result = _hidpp20.get_firmware(device)

    assert len(result) == 2
    assert isinstance(result[0], common.FirmwareInfo)
    assert isinstance(result[1], common.FirmwareInfo)


def test_get_ids():
    responses = [fake_hidpp.Response("FF12345678000D123456789ABC", 0x0400)]
    device = fake_hidpp.Device(responses=responses, feature=SupportedFeature.DEVICE_FW_VERSION)

    unitId, modelId, tid_map = _hidpp20.get_ids(device)

    assert unitId == "12345678"
    assert modelId == "123456789ABC"
    assert tid_map == {"btid": "1234", "wpid": "5678", "usbid": "9ABC"}


def test_get_kind():
    responses = [fake_hidpp.Response("00", 0x0420)]
    device = fake_hidpp.Device(responses=responses, feature=SupportedFeature.DEVICE_NAME)

    result = _hidpp20.get_kind(device)

    assert result == "keyboard"
    assert result == 1


def test_get_name():
    responses = [
        fake_hidpp.Response("12", 0x0400),
        fake_hidpp.Response("4142434445464748494A4B4C4D4E4F", 0x0410, "00"),
        fake_hidpp.Response("505152530000000000000000000000", 0x0410, "0F"),
    ]
    device = fake_hidpp.Device(responses=responses, feature=SupportedFeature.DEVICE_NAME)

    result = _hidpp20.get_name(device)

    assert result == "ABCDEFGHIJKLMNOPQR"


def test_get_friendly_name():
    responses = [
        fake_hidpp.Response("12", 0x0400),
        fake_hidpp.Response("004142434445464748494A4B4C4D4E", 0x0410, "00"),
        fake_hidpp.Response("0E4F50515253000000000000000000", 0x0410, "0E"),
    ]
    device = fake_hidpp.Device(responses=responses, feature=SupportedFeature.DEVICE_FRIENDLY_NAME)

    result = _hidpp20.get_friendly_name(device)

    assert result == "ABCDEFGHIJKLMNOPQR"


def test_get_battery_status():
    responses = [fake_hidpp.Response("502000FFFF", 0x0400)]
    device = fake_hidpp.Device(responses=responses, feature=SupportedFeature.BATTERY_STATUS)

    feature, battery = _hidpp20.get_battery_status(device)

    assert feature == SupportedFeature.BATTERY_STATUS
    assert battery.level == 80
    assert battery.next_level == 32
    assert battery.status == common.BatteryStatus.DISCHARGING


def test_get_battery_voltage():
    responses = [fake_hidpp.Response("1000FFFFFF", 0x0400)]
    device = fake_hidpp.Device(responses=responses, feature=SupportedFeature.BATTERY_VOLTAGE)

    feature, battery = _hidpp20.get_battery_voltage(device)

    assert feature == SupportedFeature.BATTERY_VOLTAGE
    assert battery.level == 92
    assert common.BatteryStatus.RECHARGING in battery.status
    assert battery.voltage == 0x1000


def test_get_battery_unified():
    responses = [fake_hidpp.Response("500100FFFF", 0x0410)]
    device = fake_hidpp.Device(responses=responses, feature=SupportedFeature.UNIFIED_BATTERY)

    feature, battery = _hidpp20.get_battery_unified(device)

    assert feature == SupportedFeature.UNIFIED_BATTERY
    assert battery.level == 80
    assert battery.status == common.BatteryStatus.DISCHARGING


def test_get_adc_measurement():
    responses = [fake_hidpp.Response("100003", 0x0400)]
    device = fake_hidpp.Device(responses=responses, feature=SupportedFeature.ADC_MEASUREMENT)

    feature, battery = _hidpp20.get_adc_measurement(device)

    assert feature == SupportedFeature.ADC_MEASUREMENT
    assert battery.level == 92
    assert battery.status == common.BatteryStatus.RECHARGING
    assert battery.voltage == 0x1000


def test_get_battery():
    responses = [fake_hidpp.Response("502000FFFF", 0x0400)]
    device = fake_hidpp.Device(responses=responses, feature=SupportedFeature.BATTERY_STATUS)

    feature, battery = _hidpp20.get_battery(device, SupportedFeature.BATTERY_STATUS)

    assert feature == SupportedFeature.BATTERY_STATUS
    assert battery.level == 80
    assert battery.next_level == 32
    assert battery.status == common.BatteryStatus.DISCHARGING


def test_get_battery_none():
    responses = [
        fake_hidpp.Response(None, 0x0000, f"{int(SupportedFeature.BATTERY_STATUS):0>4X}"),
        fake_hidpp.Response(None, 0x0000, f"{int(SupportedFeature.BATTERY_VOLTAGE):0>4X}"),
        fake_hidpp.Response("500100ffff", 0x0410),
    ]
    device = fake_hidpp.Device(responses=responses, feature=SupportedFeature.UNIFIED_BATTERY)

    feature, battery = _hidpp20.get_battery(device, None)

    assert feature == SupportedFeature.UNIFIED_BATTERY
    assert battery.level == 80
    assert battery.status == common.BatteryStatus.DISCHARGING


# get_keys is in test_hidpp20_complex
# get_remap_keys is in test_hidpp20_complex
# TODO get_gestures is complex
# get_backlight is in test_hidpp20_complex
# get_profiles is in test_hidpp20_complex


def test_get_mouse_pointer_info():
    responses = [fake_hidpp.Response("01000A", 0x0400)]
    device = fake_hidpp.Device(responses=responses, feature=SupportedFeature.MOUSE_POINTER)

    result = _hidpp20.get_mouse_pointer_info(device)

    assert result == {
        "dpi": 0x100,
        "acceleration": "med",
        "suggest_os_ballistics": False,
        "suggest_vertical_orientation": True,
    }


def test_get_vertical_scrolling_info():
    responses = [fake_hidpp.Response("01080C", 0x0400)]
    device = fake_hidpp.Device(responses=responses, feature=SupportedFeature.VERTICAL_SCROLLING)

    result = _hidpp20.get_vertical_scrolling_info(device)

    assert result == {"roller": "standard", "ratchet": 8, "lines": 12}


def test_get_hi_res_scrolling_info():
    responses = [fake_hidpp.Response("0102", 0x0400)]
    device = fake_hidpp.Device(responses=responses, feature=SupportedFeature.HI_RES_SCROLLING)

    mode, resolution = _hidpp20.get_hi_res_scrolling_info(device)

    assert mode == 1
    assert resolution == 2


def test_get_pointer_speed_info():
    responses = [fake_hidpp.Response("0102", 0x0400)]
    device = fake_hidpp.Device(responses=responses, feature=SupportedFeature.POINTER_SPEED)

    result = _hidpp20.get_pointer_speed_info(device)

    assert result == 0x0102 / 256


def test_get_lowres_wheel_status():
    responses = [fake_hidpp.Response("01", 0x0400)]
    device = fake_hidpp.Device(responses=responses, feature=SupportedFeature.LOWRES_WHEEL)

    result = _hidpp20.get_lowres_wheel_status(device)

    assert result == "HID++"


def test_get_hires_wheel():
    responses = [
        fake_hidpp.Response("010C", 0x0400),
        fake_hidpp.Response("05FF", 0x0410),
        fake_hidpp.Response("03FF", 0x0430),
    ]
    device = fake_hidpp.Device(responses=responses, feature=SupportedFeature.HIRES_WHEEL)

    multi, has_invert, has_ratchet, inv, res, target, ratchet = _hidpp20.get_hires_wheel(device)

    assert multi == 1
    assert has_invert is True
    assert has_ratchet is True
    assert inv is True
    assert res is False
    assert target is True
    assert ratchet is True


def test_get_new_fn_inversion():
    responses = [fake_hidpp.Response("0300", 0x0400)]
    device = fake_hidpp.Device(responses=responses, feature=SupportedFeature.NEW_FN_INVERSION)

    result = _hidpp20.get_new_fn_inversion(device)

    assert result == (True, False)


@pytest.fixture
def mock_gethostname(mocker):
    mocker.patch("socket.gethostname", return_value="ABCDEFG.foo.org")


@pytest.mark.parametrize(
    "responses, expected_result",
    [
        ([fake_hidpp.Response(None, 0x0400)], {}),
        ([fake_hidpp.Response("02000000", 0x0400)], {}),
        (
            [
                fake_hidpp.Response("03000200", 0x0400),
                fake_hidpp.Response("FF01FFFF05FFFF", 0x0410, "00"),
                fake_hidpp.Response("0000414243444500FFFFFFFFFF", 0x0430, "0000"),
                fake_hidpp.Response("FF01FFFF10FFFF", 0x0410, "01"),
                fake_hidpp.Response("01004142434445464748494A4B4C4D", 0x0430, "0100"),
                fake_hidpp.Response("01134E4F5000FFFFFFFFFFFFFFFFFF", 0x0430, "010E"),
                fake_hidpp.Response("000000000008", 0x0410, "00"),
                fake_hidpp.Response("0208", 0x0440, "000041424344454647"),
            ],
            {0: (True, "ABCDEFG"), 1: (True, "ABCDEFGHIJKLMNO")},
        ),
    ],
)
def test_get_host_names(responses, expected_result, mock_gethostname):
    device = fake_hidpp.Device(responses=responses, feature=SupportedFeature.HOSTS_INFO)

    result = _hidpp20.get_host_names(device)

    assert result == expected_result


@pytest.mark.parametrize(
    "responses, expected_result",
    [
        ([fake_hidpp.Response(None, 0x0400)], None),
        (
            [
                fake_hidpp.Response("03000002", 0x0400),
                fake_hidpp.Response("000000000008", 0x0410, "02"),
                fake_hidpp.Response("020E", 0x0440, "02004142434445464748494A4B4C4D4E"),
            ],
            True,
        ),
        (
            [
                fake_hidpp.Response("03000002", 0x0400),
                fake_hidpp.Response("000000000014", 0x0410, "02"),
                fake_hidpp.Response("020E", 0x0440, "02004142434445464748494A4B4C4D4E"),
                fake_hidpp.Response("0214", 0x0440, "020E4F505152535455565758"),
            ],
            True,
        ),
    ],
)
def test_set_host_name(responses, expected_result):
    device = fake_hidpp.Device(responses=responses, feature=SupportedFeature.HOSTS_INFO)

    result = _hidpp20.set_host_name(device, "ABCDEFGHIJKLMNOPQRSTUVWX")

    assert result == expected_result


def test_get_onboard_mode():
    responses = [fake_hidpp.Response("03FFFFFFFF", 0x0420)]
    device = fake_hidpp.Device(responses=responses, feature=SupportedFeature.ONBOARD_PROFILES)

    result = _hidpp20.get_onboard_mode(device)

    assert result == 0x3


def test_set_onboard_mode():
    responses = [fake_hidpp.Response("03FFFFFFFF", 0x0410, "03")]
    device = fake_hidpp.Device(responses=responses, feature=SupportedFeature.ONBOARD_PROFILES)

    res = _hidpp20.set_onboard_mode(device, 0x3)

    assert res is not None


@pytest.mark.parametrize(
    "responses, expected_result",
    [
        ([fake_hidpp.Response("03FFFF", 0x0420)], "1ms"),
        (
            [
                fake_hidpp.Response(None, 0x0000, f"{int(SupportedFeature.REPORT_RATE):04X}"),
                fake_hidpp.Response("04FFFF", 0x0420),
            ],
            "500us",
        ),
    ],
)
def test_get_polling_rate(
    responses,
    expected_result,
):
    device = fake_hidpp.Device(responses=responses, feature=SupportedFeature.EXTENDED_ADJUSTABLE_REPORT_RATE)

    result = _hidpp20.get_polling_rate(device)

    assert result == expected_result


def test_get_remaining_pairing():
    responses = [fake_hidpp.Response("03FFFF", 0x0400)]
    device = fake_hidpp.Device(responses=responses, feature=SupportedFeature.REMAINING_PAIRING)

    result = _hidpp20.get_remaining_pairing(device)

    assert result == 0x03


def test_config_change():
    responses = [fake_hidpp.Response("03FFFF", 0x0410, "02")]
    device = fake_hidpp.Device(responses=responses, feature=SupportedFeature.CONFIG_CHANGE)

    result = _hidpp20.config_change(device, 0x2)

    assert result == bytes.fromhex("03FFFF")


def test_decipher_battery_status():
    report = b"\x50\x20\x00\xff\xff"

    feature, battery = hidpp20.decipher_battery_status(report)

    assert feature == SupportedFeature.BATTERY_STATUS
    assert battery.level == 80
    assert battery.next_level == 32
    assert battery.status == common.BatteryStatus.DISCHARGING


def test_decipher_battery_voltage():
    report = b"\x10\x00\xff\xff\xff"

    feature, battery = hidpp20.decipher_battery_voltage(report)

    assert feature == SupportedFeature.BATTERY_VOLTAGE
    assert battery.level == 92
    assert common.BatteryStatus.RECHARGING in battery.status
    assert battery.voltage == 0x1000


def test_decipher_battery_unified():
    report = b"\x50\x01\x00\xff\xff"

    feature, battery = hidpp20.decipher_battery_unified(report)

    assert feature == SupportedFeature.UNIFIED_BATTERY
    assert battery.level == 80
    assert battery.status == common.BatteryStatus.DISCHARGING


def test_decipher_adc_measurement():
    report = b"\x10\x00\x03"

    feature, battery = hidpp20.decipher_adc_measurement(report)

    assert feature == SupportedFeature.ADC_MEASUREMENT
    assert battery.level == 92
    assert battery.status == common.BatteryStatus.RECHARGING
    assert battery.voltage == 0x1000


@pytest.mark.parametrize(
    "code, expected_flags",
    [
        (0x01, ["unknown:000001"]),
        (0x0F, ["unknown:00000F"]),
        (0xF0, ["internal", "hidden", "obsolete", "unknown:000010"]),
        (0x20, ["internal"]),
        (0x33, ["internal", "unknown:000013"]),
        (0x3F, ["internal", "unknown:00001F"]),
        (0x40, ["hidden"]),
        (0x50, ["hidden", "unknown:000010"]),
        (0x5F, ["hidden", "unknown:00001F"]),
        (0x7F, ["internal", "hidden", "unknown:00001F"]),
        (0x80, ["obsolete"]),
        (0xA0, ["internal", "obsolete"]),
        (0xE0, ["internal", "hidden", "obsolete"]),
        (0xFF, ["internal", "hidden", "obsolete", "unknown:00001F"]),
    ],
)
def test_feature_flag_names(code, expected_flags):
    flags = common.flag_names(hidpp20_constants.FeatureFlag, code)

    assert list(flags) == expected_flags


@pytest.mark.parametrize(
    "code, expected_name",
    [
        (0x00, "Unknown Location"),
        (0x03, "Left Side"),
    ],
)
def test_led_zone_locations(code, expected_name):
    assert hidpp20.LEDZoneLocations[code] == expected_name


@pytest.mark.parametrize(
    "millivolt, expected_percentage",
    [
        (-1234, 0),
        (500, 0),
        (2000, 0),
        (3500, 0),
        (3519, 0),
        (3520, 1),
        (3559, 1),
        (3579, 2),
        (3646, 5),
        (3671, 10),
        (3717, 20),
        (3751, 30),
        (3778, 40),
        (3811, 50),
        (3859, 60),
        (3922, 70),
        (3989, 80),
        (4067, 90),
        (4180, 99),
        (4181, 100),
        (4186, 100),
        (4500, 100),
    ],
)
def test_estimate_battery_level_percentage(millivolt, expected_percentage):
    percentage = hidpp20.estimate_battery_level_percentage(millivolt)

    assert percentage == expected_percentage
